import numpy as np
import cv2 as cv
import os
from sklearn.neighbors import LocalOutlierFactor


def splitfn(fn: str):
    """
    Splits the given file path into directory, file name, and extension.

    :param fn: file path
    :return: directory path, file name, extension
    """
    fn = os.path.abspath(fn)
    path, fn = os.path.split(fn)
    name, ext = os.path.splitext(fn)
    return path, name, ext


def smooth(x, window_len=11, window="hanning"):
    """
    Smooths the data using a window with the requested size.

    :param x: the input signal
    :param window_len: the dimension of the smoothing window; should be an odd integer
    :param window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
                   flat window will produce a moving average smoothing.
    :return: the smoothed signal

    Example:
    ```python
    t = np.linspace(-2, 2, 0.1)
    x = np.sin(t) + np.random.randn(len(t)) * 0.1
    y = smooth(x)
    ```
    """

    x = np.array(x)
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if not window in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
        raise ValueError(
            "Window is one of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"
        )

    s = np.r_[x[window_len - 1 : 0 : -1], x, x[-2 : -window_len - 1 : -1]]

    if window == "flat":  # moving average
        w = np.ones(window_len, "d")
    else:
        w = eval("np." + window + "(window_len)")

    y = np.convolve(w / w.sum(), s, mode="valid")
    return y[int(window_len / 2) - 1 : -int(window_len / 2) - 1]


def clahe_normalize(bgr):
    """
    Applies Contrast Limited Adaptive Histogram Equalization (CLAHE) normalization to the given BGR image.

    :param bgr: BGR image
    :return: normalized BGR image
    """
    lab = cv.cvtColor(bgr, cv.COLOR_BGR2LAB)
    lab[:, :, 0] = cv.createCLAHE(clipLimit=2.0, tileGridSize=(9, 9)).apply(
        lab[:, :, 0]
    )
    bgr = cv.cvtColor(lab, cv.COLOR_LAB2BGR)
    return bgr


def annotateImage(orig, flags, top=True, left=True):
    """
    Annotates the original image with flags indicating overexposure and underexposure.

    :param orig: original image
    :param flags: dictionary of flags indicating overexposure and underexposure
    :param top: boolean, whether to annotate on the top of the image
    :param left: boolean, whether to annotate on the left side of the image
    """

    try:
        y, x, _ = np.shape(orig)

        # Adjust vertical position based on 'top' argument
        if top:
            yp = int(y * 0.035)  # Place text near the top
        else:
            yp = int(y * 0.85)  # Place text near the bottom

        # Adjust horizontal position based on 'left' argument
        if left:
            xp = int(x * 0.035)  # Place text near the left
        else:
            xp = int(x * 0.85)  # Place text near the right

        # Calculate the font properties based on the image width and height
        font_scale = max(0.5, x / 2000)  # Scale font size to ensure readability
        font_thickness = max(1, x // 1000)  # Scale thickness based on image width
        offset = 0

        # Loop through flags and annotate if True
        for key in ["OVEREXPOSED", "UNDEREXPOSED"]:
            if flags.get(key, False):
                # Ensure text is within bounds of the image
                text = f"{key}: {True}"
                (text_width, text_height), _ = cv.getTextSize(
                    text, cv.FONT_HERSHEY_SIMPLEX, font_scale, font_thickness
                )

                # Adjust position if the text goes beyond the image boundaries
                if yp + offset + text_height > y:
                    yp = y - text_height - 10  # Keep text within bottom bounds

                if xp + text_width > x:
                    xp = x - text_width - 10  # Keep text within right bounds

                # Draw text on the image
                cv.putText(
                    orig,  # numpy array on which text is written
                    text,  # text
                    (xp, yp + offset),  # position at which writing has to start
                    cv.FONT_HERSHEY_SIMPLEX,  # font family
                    font_scale,  # font size
                    (0, 255, 255),  # font color (yellow in BGR)
                    font_thickness,  # font stroke
                )
            offset += int(y * 0.035)
    except Exception as e:
        print(f"Annotation error: {e}")


def getOutlierMask(metrics):
    """
    Computes the outlier mask using Local Outlier Factor (LOF).

    :param metrics: metrics data
    :return: outlier mask
    """

    X = np.nan_to_num(np.array(metrics).T)
    clf = LocalOutlierFactor(n_neighbors=9, contamination="auto")
    # use fit_predict to compute the predicted labels of the training samples
    # (when LOF is used for outlier detection, the estimator has no predict,
    # decision_function and score_samples methods).
    return clf.fit_predict(X)


def annotate_image_with_frame_number(image, frame_number):
    """
    Annotates the given image with the frame number.

    :param image: image to annotate
    :param frame_number: frame number to annotate
    """

    # Convert frame number to string
    frame_text = f"Frame: {frame_number}"

    # Get the dimensions of the image
    height, width, _ = image.shape

    # Set the font properties
    font = cv.FONT_HERSHEY_SIMPLEX
    font_scale = max(
        1, width / 1000
    )  # Scale font size to ensure readability for different image sizes
    font_thickness = max(1, width // 200)  # Scale thickness based on image width

    # Calculate the size of the frame text
    (text_width, text_height), _ = cv.getTextSize(
        frame_text, font, font_scale, font_thickness
    )

    # Calculate the position to place the frame text in the top right corner
    # Ensuring the text is within the image's height
    vertical_position = max(
        20, text_height + 10
    )  # Minimum distance from top to avoid clipping
    text_position = (width - text_width - 10, vertical_position)

    # Draw the frame text on the image
    cv.putText(
        image, frame_text, text_position, font, font_scale, (0, 255, 0), font_thickness
    )



import json
import threading
import numpy as np
import datetime
from arcjetCV.utils.utils import splitfn


class NumpyEncoder(json.JSONEncoder):
    '''
    JSON encoder for NumPy arrays.

    This class extends the JSONEncoder class to handle NumPy arrays, converting them to lists before encoding.

    Example:
    ```python
    encoder = NumpyEncoder()
    encoded_array = encoder.encode(numpy_array)
    ```

    Attributes:
        None
    '''

    def default(self, obj):
        """
        Convert NumPy arrays to lists for JSON serialization.

        :param obj: Object to encode.
        :returns: Encoded object.
        """
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, datetime.datetime):
            return obj.__str__()
        return json.JSONEncoder.default(self, obj)


class OutputListJSON(list):
    '''
    Extension of list with write to file function, expected to hold dictionary objects corresponding to analysis of individual video frames.

    This class extends the list class and provides a method to write its contents to a JSON file. It is designed to hold dictionaries corresponding to the analysis of individual video frames. The filename must contain low and high index constraints delimited by underscores, e.g., "myoutput_0_10.json".

    Args:
        list (list): Base list class.
        filepath (str): Path for saving file.

    Example:
    ```python
    output_list = OutputListJSON('output.json')
    output_list.append({'INDEX': 0, 'data': 'some_data'})
    output_list.write()
    ```

    Attributes:
        path (str): Path to the JSON file.
        folder (str): Directory containing the JSON file.
        _lock (threading.Lock): Threading lock for thread-safe operations.
        prefix (list): Prefix extracted from the filename.
        low_index (int): Low index constraint extracted from the filename.
        high_index (int): High index constraint extracted from the filename.
    '''

    def __init__(self,path):
        """
        Initializes the OutputListJSON object.

        :param path: Path for saving file.
        """
        super(OutputListJSON,self).__init__()
        self.path=path
        folder, name, _ = splitfn(path)
        self.folder = folder
        self._lock = threading.Lock()

        namesplit = name.split('_')
        self.prefix = namesplit[0:-2]
        self.low_index = int(namesplit[-2])
        self.high_index = int(namesplit[-1])

    def write(self, indent=None):
        """
        Writes the list to a JSON file.

        :param indent: Number of spaces for indentation (default=None).
        """
        with self._lock:
            json_object = json.dumps(self, indent=indent, cls=NumpyEncoder)
            # Writing to sample.json
            with open(self.path, "w") as outfile:
                outfile.write(json_object)

            print("\n\nEdges output written to", self.path)

    def load(self,path,extend=True):
        """
        Loads data from a JSON file.

        :param path: Path to the JSON file.
        :param extend: Whether to extend the current list with the loaded data (default=True).
        :returns: Loaded data.
        """
        with self._lock:
            with open(path,'r') as fin:
                dload = json.load(fin)
                if extend:
                    self.extend(dload)
                return dload

    def append(self,obj):
        """
        Appends an object to the list if its INDEX is within the specified range.

        :param obj: Object to append.
        """
        with self._lock:
            if obj["INDEX"] <= self.high_index and obj["INDEX"] >= self.low_index:
                super(OutputListJSON,self).append(obj)



import os
import urllib.request


def download_large_file(url, local_filename):
    """Downloads the file from the specified URL if it doesn't already exist locally."""
    if not os.path.exists(local_filename):
        print(f"Downloading {url} to {local_filename}...")
        urllib.request.urlretrieve(url, local_filename)
        print(f"Downloaded {local_filename}.")
    else:
        print(f"{local_filename} already exists, skipping download.")


def download_test_video():
    """Ensures that the test video 'arcjet_test.mp4' is downloaded."""
    base_url = "https://github.com/magnus-haw/arcjetCV/raw/main/arcjetCV/tests/"
    file_to_download = "arcjet_test.mp4"

    download_large_file(
        base_url + file_to_download,
        os.path.join(os.path.dirname(__file__), file_to_download),
    )


# Automatically download the video when the function is called
download_test_video()



import cv2
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from scipy import interpolate
from pathlib import Path


def _torch_load_compat(path, **kwargs):
    try:
        return torch.load(path, weights_only=False, **kwargs)
    except TypeError:
        return torch.load(path, **kwargs)


def extract_interest(signal):
    """Extract the starts and ends of a signal that is the mean brightness of frames
    Args:
        signal (array): mean brightness signal of frames
    Returns:
        arrays: start and end indexes
    """
    start = []
    end = []
    for index, i in enumerate(signal):
        if index == 0 and i > 0.5:
            start.append(index)
        elif signal[index - 1] <= 0.5 and i > 0.5:
            start.append(index)
        elif signal[index - 1] >= 0.5 and i < 0.5:
            end.append(index)
    if signal[-1] > 0.5:
        end.append(len(signal) - 1)
    return start, end


def time_segmentation(video, progress_callback=None):
    """apply a segmentation on a the mean brightness of frames of a video
    """
    max_sampled_frames = 120
    baseline_samples = max(2, ((video.nframes - 1) // 10) + 1)
    nsamples = min(max_sampled_frames, baseline_samples)
    sample_indices = np.linspace(0, video.nframes - 1, num=nsamples, dtype=int)

    nFrame = []
    Value = []
    if progress_callback is not None:
        progress_callback(0, "Detecting sample insertion")
    for frame_index in sample_indices:
        try:
            _frame = video.get_frame(int(frame_index))
            h = _frame.shape[0]
            y0 = int(0.2 * h)
            y1 = int(0.8 * h)
            roi = _frame[y0:y1, :]
            if roi.size == 0:
                continue

            # Downsample before computing brightness to speed up loading on large frames.
            max_dim = 256
            if max(roi.shape[0], roi.shape[1]) > max_dim:
                scale = max_dim / float(max(roi.shape[0], roi.shape[1]))
                roi = cv2.resize(
                    roi,
                    (
                        max(1, int(roi.shape[1] * scale)),
                        max(1, int(roi.shape[0] * scale)),
                    ),
                    interpolation=cv2.INTER_AREA,
                )

            # Keep feature extraction aligned with the checkpoint training pipeline:
            # erode + HSV V-channel brightness.
            kernel = np.ones((20, 20), np.uint8)
            roi = cv2.erode(roi, kernel, cv2.BORDER_REFLECT)
            hsv = cv2.cvtColor(roi, cv2.COLOR_RGB2HSV)
            Value.append(np.mean(hsv[:, :, 2]))
            nFrame.append((frame_index / max(1, video.nframes - 1)) * 500.0)
            if progress_callback is not None:
                progress = int((len(Value) / max(1, nsamples)) * 100)
                progress_callback(progress, "Detecting sample insertion...")
        except Exception:
            pass

    if len(Value) < 2:
        raise RuntimeError("Insufficient sampled frames for time segmentation.")

    Value = np.array(Value, dtype=float)
    value_range = np.ptp(Value)
    if value_range <= 0:
        raise RuntimeError("Brightness signal is constant; cannot segment in time.")
    Value = (Value - np.min(Value)) / value_range

    nFrame = np.array(nFrame, dtype=float)
    p = interpolate.interp1d(nFrame, Value, fill_value="extrapolate")
    reg = np.linspace(0, 500, 500)

    model = Conv1DNet()
    checkpoint_path = Path(__file__).parent.absolute().joinpath(Path('time-checkpoint.pt'))
    model.load_state_dict(_torch_load_compat(checkpoint_path))
    if progress_callback is not None:
        progress_callback(100, "Detecting sample insertion")
    input = np.array(p(reg))

    # Prepare input tensor with shape [1, 1, 500]
    input = torch.tensor(input, dtype=torch.float).unsqueeze(0).unsqueeze(1)
    model.eval()  # Switch to evaluation mode
    with torch.no_grad():  # Disable gradient computation
        test_output = model(input)
        probabilities = F.softmax(
            test_output, dim=1
        )  # Optional: convert logits to probabilities
        predictions = torch.argmax(probabilities, dim=1).squeeze()  # Predicted classes

    # Flatten input and probabilities for plotting and analysis
    input = input.squeeze().cpu().numpy()
    out = predictions.squeeze().cpu().numpy()

    return input, out


class Conv1DNet(nn.Module):
    def __init__(self):
        super(Conv1DNet, self).__init__()
        self.conv1 = nn.Conv1d(
            in_channels=1, out_channels=32, kernel_size=7, stride=2, padding=3
        )
        self.conv2 = nn.Conv1d(
            in_channels=32, out_channels=16, kernel_size=7, stride=2, padding=3
        )
        self.convtrans1 = nn.ConvTranspose1d(
            in_channels=16,
            out_channels=16,
            kernel_size=7,
            stride=2,
            padding=3,
            output_padding=1,
        )
        self.convtrans2 = nn.ConvTranspose1d(
            in_channels=16,
            out_channels=32,
            kernel_size=7,
            stride=2,
            padding=3,
            output_padding=1,
        )
        self.convtrans3 = nn.ConvTranspose1d(
            in_channels=32, out_channels=3, kernel_size=3, stride=1, padding=1
        )
        self.dropout = nn.Dropout(0.25)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = self.dropout(x)
        x = F.relu(self.conv2(x))
        x = self.dropout(x)
        x = F.relu(self.convtrans1(x))
        x = self.dropout(x)
        x = F.relu(self.convtrans2(x))
        x = self.dropout(x)
        x = self.convtrans3(x)
        # Applying softmax on the last dimension (channels) to get probabilities
        x = F.softmax(x, dim=1)
        return x



"""Copyright © 2024 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.
Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS, RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS AGREEMENT.

Root directory for the arcjetCV package.
"""

# Note: only this version has to be bumped for the whole project
# to publish new version: (pre-requisite: conda install conda-forge::gh)
# - bump version below
# - git add -A
# - git commit -m 'Version bumped'
# - git push origin main
# - git tag -a v$(python setup.py --version) -m 'INPUT DESCRIPTION'
# - gh release create v$(python setup.py --version) --target main
__version__ = "1.1.17"

from arcjetCV.utils.video import Video, VideoMeta
from arcjetCV.utils.processor import ArcjetProcessor



import os
from pathlib import Path
import numpy as np
import pandas as pd
import cv2 as cv
import json
from numbers import Number
from PySide6 import QtWidgets
from PySide6.QtWidgets import QMessageBox
from PySide6.QtCore import Signal, QThread, Qt
from PySide6.QtGui import QIcon, QPixmap
import matplotlib.pyplot as plt
from matplotlib.colors import rgb_to_hsv
from matplotlib.widgets import RectangleSelector
from arcjetCV.gui.arcjetCV_gui import Ui_MainWindow
from arcjetCV.utils.video import Video, VideoMeta
from arcjetCV.utils.processor import ArcjetProcessor, ProcessorWorker
from arcjetCV.utils.utils import (
    splitfn,
    getOutlierMask,
    annotateImage,
    annotate_image_with_frame_number,
)
from arcjetCV.calibration.calibration_controller import CalibrationController
from PySide6.QtWidgets import QFileDialog
import json


class MainWindow(QtWidgets.QMainWindow):
    frame_processed = Signal()

    def __init__(self):
        super().__init__()

        # Initialize the user interface
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.testing = False

        # Set the application icon
        self.logo_white_path = os.path.join(
            Path(__file__).parent.absolute(), "logo/arcjetCV_logo_white.png"
        )
        self.logo_path = os.path.join(
            Path(__file__).parent.absolute(), "logo/arcjetCV_logo_.png"
        )
        self.setWindowIcon(QIcon(self.logo_path))

        # Load and process the application logo
        self.rgb_frame = cv.imread(self.logo_white_path)
        self.rgb_frame = cv.cvtColor(self.rgb_frame, cv.COLOR_BGR2RGB)

        # Initialize frame and plotting windows
        self.calibrated = None
        self.calibration_data = None
        self.pixels_per_mm = None
        self._plot_ref = None
        self._tplot_ref = None
        self._brightness_ref = None
        self.show_img()

        # Initialize line start and stop positions
        self.start_line = None
        self.stop_line = None

        # Flags for video loading and processing
        self.VIDEO_LOADED = False
        self.NEW_VIDEO = False

        # Initialize folder/file properties
        self.folder = None
        self.path = None
        self.filename = None
        self.ext = None
        self.video = None
        self.videometa = None

        # Processor objects
        self.processor = None

        # Data structures
        self.raw_outputs = []
        self.time_series = None
        self.PLOTKEYS = []
        self.fit_dict = None
        self._plot_ref = None

        # Connect interface buttons to their respective functions
        self.ui.pushButton_process.clicked.connect(self.process_all)
        self.ui.pushButton_loadVideo.clicked.connect(self.load_video)
        self.ui.pushButton_loadCalibration.clicked.connect(self.load_calibration)
        self.ui.pushButton_export_csv.clicked.connect(self.export_to_csv)
        self.ui.pushButton_fitData.clicked.connect(self.fit_data)
        self.ui.pushButton_LoadFiles.clicked.connect(self.load_outputs)
        self.ui.pushButton_PlotData.clicked.connect(self.plot_outputs)
        self.ui.checkBox_display_shock2.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_m95_radius.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_m50_radius.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_model_center.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_50_radius.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_95_radius.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_shock_area.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_model_rad.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_shock_center.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_shockmodel.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_ypos.stateChanged.connect(self.plot_outputs)
        self.ui.checkBox_kalman_filter.stateChanged.connect(self.plot_outputs)
        self.ui.comboBox_units.setCurrentText("[mm]")
        self._set_tooltips()

        # init_plot_brightness
        self.ui.Window3.canvas.axes.clear()
        self.ui.Window3.canvas.axes.set_yticklabels([])
        self.ui.Window3.canvas.axes.get_xaxis().set_visible(False)
        self.ui.Window3.canvas.axes.get_yaxis().set_visible(False)
        if self.ui.Window3.canvas.toolbar is not None:
            self.ui.Window3.canvas.toolbar.setVisible(False)
        self.ui.Window3.canvas.draw()

        # Set some UI elements to be initially disabled
        self.ui.spinBox_FrameIndex.setDisabled(True)
        self.ui.comboBox_flowDirection.setDisabled(True)
        self.ui.comboBox_filterType.setDisabled(True)
        self.ui.FilterTabs.setDisabled(True)
        self.ui.comboBox_filterType.setDisabled(True)
        self.ui.comboBox_filterType.setDisabled(True)

        # Show the main window
        self.show()

    def _set_tooltips(self):
        """Set hover help for key UI controls."""
        tooltips = {
            "pushButton_loadCalibration": "Load a calibration file to convert pixels to physical units.",
            "pushButton_loadVideo": "Open a video file for processing.",
            "pushButton_save_frame": "Save the currently displayed frame as an image.",
            "pushButton_process": "Process the selected frame range and extract time-series outputs.",
            "pushButton_LoadFiles": "Load previously exported output files for plotting.",
            "pushButton_export_csv": "Export current results to a CSV file.",
            "pushButton_PlotData": "Plot selected quantities from loaded or processed data.",
            "pushButton_fitData": "Fit the selected data over the chosen time interval.",
            "applyCrop": "Apply the current crop bounds to the video processing region.",
            "spinBox_FrameIndex": "Current frame index shown in the preview.",
            "comboBox_flowDirection": "Direction of the flow used by the detector.",
            "comboBox_filterType": "Segmentation method: HSV, GRAY, CNN, or AutoHSV.",
            "checkBox_display_shock": "Enable/disable shock detection overlay on the preview frame.",
            "checkBox_crop": "Show/hide crop rectangle and use crop bounds during processing.",
            "checkBox_annotate": "Overlay computed values and labels on the preview frame.",
            "spinBox_crop_xmin": "Left crop boundary (pixels).",
            "spinBox_crop_xmax": "Right crop boundary (pixels).",
            "spinBox_crop_ymin": "Top crop boundary (pixels).",
            "spinBox_crop_ymax": "Bottom crop boundary (pixels).",
            "spinBox_FirstGoodFrame": "First frame included in batch processing.",
            "spinBox_LastGoodFrame": "Last frame included in batch processing.",
            "spinBox_frame_skips": "Process every Nth frame to speed up analysis.",
            "checkBox_writeVideo": "Write an output video with overlays while processing.",
            "comboBox_units": "Unit system used in plots and derived quantities.",
            "doubleSpinBox_fps": "Frame rate used for the time axis (frames per second).",
            "spinBox_mask_frames": "Minimum number of valid points required for plotting/fitting masks.",
            "checkBox_display_shock2": "Include shock position in the time-series plot.",
            "checkBox_kalman_filter": "Apply Kalman smoothing to plotted signals.",
            "comboBox_fit_type": "Model used for fitting (currently linear).",
            "doubleSpinBox_fit_start_time": "Fit window start time (seconds).",
            "doubleSpinBox_fit_last_time": "Fit window end time (seconds).",
            "checkBox_m95_radius": "Plot model 95% radius.",
            "checkBox_m50_radius": "Plot model 50% radius.",
            "checkBox_model_center": "Plot model center position.",
            "checkBox_50_radius": "Plot measured 50% radius.",
            "checkBox_95_radius": "Plot measured 95% radius.",
            "checkBox_shock_area": "Plot shock area.",
            "checkBox_model_rad": "Plot model radius.",
            "checkBox_shock_center": "Plot shock center position.",
            "checkBox_shockmodel": "Plot shock model result.",
            "checkBox_ypos": "Plot Y-position metric.",
        }

        for name, tip in tooltips.items():
            widget = getattr(self.ui, name, None)
            if widget is not None:
                widget.setToolTip(tip)
                widget.setStatusTip(tip)

    def _set_calibration_resolution_label(self):
        """Display loaded calibration resolution as mm/px and px/mm in the UI."""
        label = getattr(self.ui, "label_calibrationResolution", None)
        if label is None:
            return

        ppm = getattr(self, "pixels_per_mm", None)
        if ppm in (None, 0):
            label.setText("Resolution: None")
            return

        mm_per_px = 1.0 / ppm
        label.setText(f"Resolution: {mm_per_px:.6f} mm/px | {ppm:.6f} px/mm")

    @staticmethod
    def _flow_axis_sign(flow_direction):
        """
        Return a sign that makes downstream motion positive across flow directions.
        """
        flow = str(flow_direction).strip().lower()
        return -1.0 if flow in {"left", "up"} else 1.0

    def closeEvent(self, event):
        print("🔴 Closing application...")
        try:
            if hasattr(self, "worker") and self.worker is not None:
                print("🛑 Attempting to stop worker...")
                self.worker.stop()
                self.worker.deleteLater()
                self.worker = None

            if hasattr(self, "thread") and isinstance(self.thread, QThread):
                if self.thread.isRunning():
                    print("🛑 Waiting for thread to finish...")
                    self.thread.quit()
                    self.thread.wait(5000)
                    print("✅ Thread fully stopped.")
                self.thread.deleteLater()
                self.thread = None

        except Exception as e:
            print(f"[ERROR] during closeEvent: {e}")

        event.accept()


    def show_img(self):
        # if self.calibrated == True:
        #     self.rgb_frame = CalibrationController.apply_calibration(
        #         self.rgb_frame, self.calibration_data
        #     )
        #     print("Calibrated")
        if self._plot_ref is None:
            # Create a new plot reference and define the cursor data format
            self._plot_ref = self.ui.Window0.canvas.axes.imshow(
                self.rgb_frame, aspect="equal"
            )
            self._plot_ref.axes.get_xaxis().set_visible(False)
            self._plot_ref.axes.get_yaxis().set_visible(False)

            # Customize axis spines color to match the background (hiding the border)
            self._plot_ref.axes.spines["top"].set_color("none")
            self._plot_ref.axes.spines["bottom"].set_color("none")
            self._plot_ref.axes.spines["left"].set_color("none")
            self._plot_ref.axes.spines["right"].set_color("none")

            def format_cursor_data(data):
                try:
                    data[0]
                except (TypeError, IndexError):
                    data = [data]
                data_str = "RGB = [" + ", ".join(
                    "{:0.3g}".format(item) for item in data if isinstance(item, Number)
                )
                hsv = rgb_to_hsv(np.array(data) / 255.0)
                hsv_str = (
                    "HSV = ["
                    + "%i " % (hsv[0] * 180)
                    + ","
                    + "%i " % (hsv[1] * 255)
                    + ","
                    + "%i " % (hsv[2] * 255)
                    + "]"
                )
                return data_str + "] " + hsv_str

            self._plot_ref.format_cursor_data = format_cursor_data
            self.rect_selector = RectangleSelector(self._plot_ref.axes, self.onselect)
        else:
            self._plot_ref.set_data(self.rgb_frame)

        self.ui.Window0.canvas.draw()

    # def onselect(self, eclick, erelease):
    #     x1, y1 = eclick.xdata, eclick.ydata
    #     x2, y2 = erelease.xdata, erelease.ydata
    #     self.ui.spinBox_crop_xmin.setValue(x1)
    #     self.ui.spinBox_crop_xmax.setValue(x2)
    #     self.ui.spinBox_crop_ymin.setValue(y1)
    #     self.ui.spinBox_crop_ymax.setValue(y2)
    #     self.update_crop()

    def onselect(self, eclick, erelease):
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata

        # If image was rotated 90° clockwise, reverse it
        # Convert (x, y) to pre-rotation coordinates
        print(self.calibrated)
        # if self.calibrated:
        #     print("good crop")
        #     h, w = self.rgb_frame.shape[:2]

        #     def unrotate_point(x, y):
        #         # Reverse 90° clockwise = rotate 90° counter-clockwise
        #         return y, w - x

        #     x1, y1 = unrotate_point(x1, y1)
        #     x2, y2 = unrotate_point(x2, y2)

        #     # Optional: undo homography if you want original pixel space
        #     H_inv = np.array(self.calibration_data["homography_inverse"])
        #     pts = np.array([[x1, y1], [x2, y2]], dtype=np.float32).reshape(-1, 1, 2)
        #     pts_original = cv.perspectiveTransform(pts, H_inv).reshape(-1, 2)
        #     x1, y1 = pts_original[0]
        #     x2, y2 = pts_original[1]

        # Store corrected coordinates
        self.ui.spinBox_crop_xmin.setValue(int(min(x1, x2)))
        self.ui.spinBox_crop_xmax.setValue(int(max(x1, x2)))
        self.ui.spinBox_crop_ymin.setValue(int(min(y1, y2)))
        self.ui.spinBox_crop_ymax.setValue(int(max(y1, y2)))
        self.update_crop()

    def update_crop(self):
        # Validate bounding box inputs here
        if (
            self.ui.spinBox_crop_xmax.value() - self.ui.spinBox_crop_xmin.value() >= 40
            and self.ui.spinBox_crop_ymax.value() - self.ui.spinBox_crop_ymin.value()
            >= 40
        ):
            self.videometa["CROP_XMIN"] = self.ui.spinBox_crop_xmin.value()
            self.videometa["CROP_XMAX"] = self.ui.spinBox_crop_xmax.value()
            self.videometa["CROP_YMIN"] = self.ui.spinBox_crop_ymin.value()
            self.videometa["CROP_YMAX"] = self.ui.spinBox_crop_ymax.value()
            self.videometa.write()
            self.update_frame_index()
        else:
            pass

    def brightness_click_slot(self, x, y):
        try:
            if x > 0 and x < self.video.nframes - 1:
                self.ui.spinBox_FrameIndex.setValue(int(x))
        except:
            pass

    def update_frame_index(self):

        if self.NEW_VIDEO:
            self.ui.spinBox_crop_xmin.setValue(self.videometa["CROP_XMIN"])
            self.ui.spinBox_crop_xmax.setValue(self.videometa["CROP_XMAX"])
            self.ui.spinBox_crop_ymin.setValue(self.videometa["CROP_YMIN"])
            self.ui.spinBox_crop_ymax.setValue(self.videometa["CROP_YMAX"])

        self.ui.maxHue.setDisabled(False)
        self.ui.minHue.setDisabled(False)
        self.ui.minIntensity.setDisabled(False)
        self.ui.maxIntensity.setDisabled(False)
        self.ui.minSaturation.setDisabled(False)
        self.ui.maxSaturation.setDisabled(False)
        self.ui.maxHue_2.setDisabled(False)
        self.ui.minHue_2.setDisabled(False)
        self.ui.minIntensity_2.setDisabled(False)
        self.ui.maxIntensity_2.setDisabled(False)
        self.ui.minSaturation_2.setDisabled(False)
        self.ui.maxSaturation_2.setDisabled(False)

        # disabled the tabs with CNN
        if (
            self.ui.comboBox_filterType.currentText() == "CNN"
            or self.ui.comboBox_filterType.currentText() == "AutoHSV"
        ):
            self.ui.tab_3.setDisabled(True)
            self.ui.tab_4.setDisabled(True)
        elif self.ui.comboBox_filterType.currentText() == "GRAY":
            self.ui.tab_3.setDisabled(False)
            self.ui.tab_4.setDisabled(False)
            self.ui.maxHue.setDisabled(True)
            self.ui.minHue.setDisabled(True)
            self.ui.minIntensity.setDisabled(False)
            self.ui.maxIntensity.setDisabled(False)
            self.ui.minSaturation.setDisabled(True)
            self.ui.maxSaturation.setDisabled(True)
            self.ui.maxHue_2.setDisabled(True)
            self.ui.minHue_2.setDisabled(True)
            self.ui.minIntensity_2.setDisabled(False)
            self.ui.maxIntensity_2.setDisabled(False)
            self.ui.minSaturation_2.setDisabled(True)
            self.ui.maxSaturation_2.setDisabled(True)
        else:
            self.ui.tab_3.setDisabled(False)
            self.ui.tab_4.setDisabled(False)

        frame_index = self.ui.spinBox_FrameIndex.value()
        self.rgb_frame = self.video.get_frame(frame_index)
        if self.rgb_frame is None:
            return
        if self.calibrated == True:
            self.rgb_frame = CalibrationController.apply_calibration(
                self.rgb_frame, self.calibration_data
            )
            print("Calibrated")

        self.processor.flow_dir = self.ui.comboBox_flowDirection.currentText()

        crop_range = [
            [self.ui.spinBox_crop_ymin.value(), self.ui.spinBox_crop_ymax.value()],
            [self.ui.spinBox_crop_xmin.value(), self.ui.spinBox_crop_xmax.value()],
        ]

        # Apply crop params
        if self.processor.crop != crop_range:
            self.processor.crop = crop_range
            if (
                self.VIDEO_LOADED
            ):  # To avoid crash (can be applied to all the function update_frame)
                self.videometa["CROP_YMIN"] = self.ui.spinBox_crop_ymin.value()
                self.videometa["CROP_XMIN"] = self.ui.spinBox_crop_xmin.value()
                self.videometa["CROP_YMAX"] = self.ui.spinBox_crop_ymax.value()
                self.videometa["CROP_XMAX"] = self.ui.spinBox_crop_xmax.value()
                self.videometa.write()

        # Process frame
        contour_dict, argdict = self.processor.process(
            self.rgb_frame, self.grab_ui_values()
        )

        # Draw contours
        width = self.rgb_frame.shape[1]
        thickness = max(
            1, width // 500
        )  # Scale thickness with width, minimum thickness of 1
        for key in contour_dict.keys():
            if key == "MODEL":
                cv.drawContours(
                    self.rgb_frame, contour_dict[key], -1, (0, 255, 0), thickness
                )
            elif key == "SHOCK" and self.ui.checkBox_display_shock.isChecked():
                cv.drawContours(
                    self.rgb_frame, contour_dict[key], -1, (255, 0, 0), thickness
                )

        # Draw annotations
        annotate_image_with_frame_number(self.rgb_frame, frame_index)
        if self.ui.checkBox_annotate.isChecked():
            annotateImage(self.rgb_frame, argdict, top=True, left=True)

        if self.ui.checkBox_crop.isChecked():
            start_point = (self.processor.crop[1][0], self.processor.crop[0][0])
            end_point = (self.processor.crop[1][1], self.processor.crop[0][1])
            cv.rectangle(
                self.rgb_frame, start_point, end_point, (255, 255, 255), thickness
            )

        self.frame_processed.emit()
        self.plot_location()

    def save_frame(self):
        frame_index = self.ui.spinBox_FrameIndex.value()
        frame = self.video.get_frame(frame_index)
        # Convert from BGR to RGB
        image = cv.cvtColor(frame, cv.COLOR_RGB2BGR)

        file_dialog = QtWidgets.QFileDialog()
        file_dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptSave)
        file_dialog.setNameFilter("Images (*.png *.jpg *.bmp *.tif)")
        try:
            if file_dialog.exec():
                fpath = file_dialog.selectedFiles()[0]
                cv.imwrite(fpath, image)
                print("------SAVED FRAME %i-----" % frame_index)
                self.ui.basebar.setText("------SAVED FRAME %i-----" % frame_index)
        except:
            self.arcjetcv_message_box("Warning", "Could not save frame")

    def connect_elements(self):
        self.ui.spinBox_FrameIndex.valueChanged.connect(self.update_frame_index)
        self.frame_processed.connect(self.show_img)
        self.ui.checkBox_display_shock.stateChanged.connect(self.update_frame_index)
        self.ui.maxHue.valueChanged.connect(self.update_frame_index)
        self.ui.minHue.valueChanged.connect(self.update_frame_index)
        self.ui.minIntensity.valueChanged.connect(self.update_frame_index)
        self.ui.maxIntensity.valueChanged.connect(self.update_frame_index)
        self.ui.minSaturation.valueChanged.connect(self.update_frame_index)
        self.ui.maxSaturation.valueChanged.connect(self.update_frame_index)
        self.ui.maxHue_2.valueChanged.connect(self.update_frame_index)
        self.ui.minHue_2.valueChanged.connect(self.update_frame_index)
        self.ui.minIntensity_2.valueChanged.connect(self.update_frame_index)
        self.ui.maxIntensity_2.valueChanged.connect(self.update_frame_index)
        self.ui.minSaturation_2.valueChanged.connect(self.update_frame_index)
        self.ui.maxSaturation_2.valueChanged.connect(self.update_frame_index)
        self.ui.applyCrop.clicked.connect(self.update_crop)
        self.ui.pushButton_save_frame.clicked.connect(self.save_frame)
        self.ui.checkBox_crop.stateChanged.connect(self.update_frame_index)
        if (
            self.ui.spinBox_crop_xmin.value() < self.ui.spinBox_crop_xmax.value()
            and self.ui.spinBox_crop_ymin.value() < self.ui.spinBox_crop_ymax.value()
        ):
            self.ui.applyCrop.clicked.connect(self.update_frame_index)
        self.ui.comboBox_filterType.currentTextChanged.connect(self.update_frame_index)
        self.ui.comboBox_flowDirection.currentTextChanged.connect(
            self.update_frame_index
        )

    def load_video(self):
        print("------New Video-----")
        dialog = QtWidgets.QFileDialog()
        options = QtWidgets.QFileDialog.Options()
        pathmask = dialog.getOpenFileName(
            None,
            "Select Video",
            "",
            "Video Files (*.mp4 *.avi *.mov *.m4v);;All Files (*)",
            options=options,
        )
        self.NEW_VIDEO = True

        self.path = pathmask[0]
        self._plot_ref = None
        if self.path != "":
            self.folder, self.filename, self.ext = splitfn(self.path)
            loading_dialog = None

            if not self.testing:
                loading_dialog = QtWidgets.QProgressDialog(
                    "Loading video...", "Cancel", 0, 100, self
                )
                loading_dialog.setWindowTitle("Loading Video")
                loading_dialog.setWindowModality(Qt.WindowModal)
                loading_dialog.setMinimumDuration(0)
                loading_dialog.setAutoClose(False)
                loading_dialog.setValue(0)
                loading_dialog.show()

            def update_loading_progress(value, message):
                if loading_dialog is None:
                    return
                loading_dialog.setValue(int(max(0, min(100, value))))
                if message:
                    loading_dialog.setLabelText(message)
                QtWidgets.QApplication.processEvents()
                if loading_dialog.wasCanceled():
                    raise RuntimeError("Video loading canceled by user.")

            # Old matplotlib artists become invalid after axes.clear(); drop refs.
            self.start_line = None
            self.stop_line = None
            self._tplot_ref = None

            self.ui.spinBox_FrameIndex.setDisabled(False)
            self.ui.comboBox_flowDirection.setDisabled(False)
            self.ui.comboBox_filterType.setDisabled(False)
            self.ui.FilterTabs.setDisabled(False)
            self.ui.comboBox_filterType.setDisabled(False)
            self.ui.comboBox_filterType.setDisabled(False)
            self.ui.Window0.canvas.axes.clear()
            self.ui.Window0.canvas.draw()

            try:  # Create video object
                update_loading_progress(2, "Opening video file...")
                self.video = Video(self.path)
                self.videometa = VideoMeta(
                    self.video,
                    os.path.join(self.folder, self.filename + ".meta"),
                    progress_callback=update_loading_progress,
                )

                if hasattr(self, "pixels_per_mm"):  # Ensure it's defined
                    self.videometa["PIXELS_PER_MM"] = self.pixels_per_mm
                    print(
                        f"DEBUG: Stored pixels_per_mm in videometa: {self.videometa['PIXELS_PER_MM']}"
                    )  # Debugging
                else:
                    print("WARNING: pixels_per_mm not set before loading video.")
                print("Number of Frames: ", self.video.nframes)
                print(
                    f"First-Last Good Frame: {self.videometa['FIRST_GOOD_FRAME']}-{self.videometa['LAST_GOOD_FRAME']}"
                )
                print(
                    f"Min-Max Brightness: {min(self.videometa['BRIGHTNESS'])}-{max(self.videometa['BRIGHTNESS'])}"
                )

                # Initialize UI
                self.ui.Window3.canvas.axes.clear()
                self.ax1 = self.ui.Window3.canvas.axes
                self.ui.spinBox_FrameIndex.setRange(0, self.video.nframes - 1)
                self.ui.spinBox_FrameIndex.setValue(self.videometa["FIRST_GOOD_FRAME"])
                self.ui.spinBox_FirstGoodFrame.setValue(
                    self.videometa["FIRST_GOOD_FRAME"]
                )
                self.ui.spinBox_LastGoodFrame.setValue(
                    self.videometa["LAST_GOOD_FRAME"]
                )
                self.ui.comboBox_flowDirection.setCurrentText(
                    self.videometa["FLOW_DIRECTION"]
                )
                self.ui.lineEdit_filename.setText(self.video.name)
                self.ui.spinBox_LastGoodFrame.setRange(0, self.video.nframes - 1)
                self.ui.spinBox_FirstGoodFrame.setRange(0, self.video.nframes - 1)

                # Plot the brightness signal
                self.ax1.plot(self.videometa["BRIGHTNESS"])
                self.ax1.autoscale_view()
                self.ax1.set_xlim([0, self.video.nframes - 1])

                # Plot the red point for the specific n value and start and stop
                self.plot_location(reset=True)
                self.plot_start_stop()

                self.ui.Window3.canvas.draw()

                self.ui.Window3.canvas.clicked.connect(
                    self.brightness_click_slot, Qt.UniqueConnection
                )
                self.ui.spinBox_FirstGoodFrame.valueChanged.connect(
                    self.plot_start_stop, Qt.UniqueConnection
                )
                self.ui.spinBox_LastGoodFrame.valueChanged.connect(
                    self.plot_start_stop, Qt.UniqueConnection
                )
                if self.processor is None:
                    update_loading_progress(99, "Initializing processor...")
                    self.processor = ArcjetProcessor(
                        self.videometa, progress_bar=self.ui.progressBar
                    )
                else:  # avoid reimporting CNN model each time
                    self.processor.update_video_meta(self.videometa)

                # Connect UI only on first video load
                if self.VIDEO_LOADED is False:
                    self.connect_elements()
                    self.VIDEO_LOADED = True

                    self.update_frame_index()
                    self.NEW_VIDEO = False
                update_loading_progress(100, "Video loaded.")

            except Exception as e:
                if self.testing:
                    print("Could not load video:\n" + str(e))
                else:
                    self.arcjetcv_message_box(
                        "Warning", "Could not load video:\n" + str(e)
                    )
            finally:
                if loading_dialog is not None:
                    loading_dialog.close()

    def load_calibration(self):
        """
        Load a JSON calibration file and store the available calibration parameters.

        Only existing keys in the JSON file will be loaded. Missing keys are skipped.
        The calibration data is stored in self.calibration_data and self.pixels_per_mm.
        """
        file_path, _ = QFileDialog.getOpenFileName(
            None, "Select Calibration File", "", "JSON Files (*.json)"
        )

        if not file_path:
            return

        try:
            with open(file_path, "r") as file:
                raw_data = json.load(file)

            # Initialize calibration data dictionary
            self.calibration_data = {}

            # Load each expected key if it exists
            key_types = {
                "camera_matrix": (np.float32, (3, 3)),
                "dist_coeffs": (np.float32, None),
                "rvec": (np.float32, None),
                "tvec": (np.float32, None),
                "pixels_per_mm": (float, None),
                "centers": (np.float32, None),
                "square_layout": (np.float32, None),
                "homography": (np.float32, None),
                "homography_inverse": (np.float32, None),
                "pattern_size": (tuple, None),
                "affine_matrix": (np.float32, None),
            }

            for key, (dtype, shape) in key_types.items():
                if key in raw_data:
                    value = raw_data[key]
                    if dtype == np.float32:
                        arr = np.array(value, dtype=dtype)
                        self.calibration_data[key] = arr
                    elif dtype == float:
                        self.calibration_data[key] = float(value)
                    elif dtype == tuple:
                        self.calibration_data[key] = tuple(value)

            # Set ppm and status flag
            self.pixels_per_mm = self.calibration_data.get("pixels_per_mm", None)
            self.calibrated = True

            # Propagate scaling immediately when a video/processor is already active.
            if self.pixels_per_mm is not None:
                if self.videometa is not None:
                    self.videometa["PIXELS_PER_MM"] = self.pixels_per_mm
                    self.videometa.write()
                if self.processor is not None:
                    self.processor.pixels_per_mm = self.pixels_per_mm

            # Update label
            shortpath = self.shorten_path(file_path, 60)
            self.ui.label_calibrationPath.setText(f"Calibration Path: {shortpath}")
            self._set_calibration_resolution_label()

            print("✅ Calibration data loaded:", list(self.calibration_data.keys()))

        except Exception as e:
            self.arcjetcv_message_box(
                "Error", f"Failed to load calibration file: {str(e)}"
            )
            self.calibration_data = None
            self.calibrated = False
            self.pixels_per_mm = None
            self._set_calibration_resolution_label()
        if self.VIDEO_LOADED:
            self.update_frame_index()

    def plot_location(self, reset=False):
        if self.video is None:
            return
        n = self.ui.spinBox_FrameIndex.value()
        current_axes = self.ui.Window3.canvas.axes
        if (
            self._tplot_ref is None
            or reset
            or getattr(self._tplot_ref, "axes", None) is not current_axes
        ):
            self._tplot_ref = self.ui.Window3.canvas.axes.axvline(
                x=n, color="red", linestyle="-"
            )
        else:
            self._tplot_ref.set_xdata([n, n])
        self.ui.Window3.canvas.draw()

    def plot_start_stop(self):
        if self.video is None:
            return

        start = self.ui.spinBox_FirstGoodFrame.value()
        stop = self.ui.spinBox_LastGoodFrame.value()
        current_axes = self.ui.Window3.canvas.axes

        if 0 <= start < self.video.nframes:
            if (
                self.start_line is None
                or getattr(self.start_line, "axes", None) is not current_axes
            ):
                self.start_line = current_axes.axvline(
                    x=start, color="green", linestyle="-"
                )
            else:
                self.start_line.set_xdata([start, start])
        else:
            if self.start_line is not None:
                try:
                    self.start_line.remove()
                except (NotImplementedError, ValueError):
                    pass
            self.start_line = None

        if 0 <= stop < self.video.nframes:
            if (
                self.stop_line is None
                or getattr(self.stop_line, "axes", None) is not current_axes
            ):
                self.stop_line = current_axes.axvline(
                    x=stop, color="green", linestyle="-"
                )
            else:
                self.stop_line.set_xdata([stop, stop])
        else:
            if self.stop_line is not None:
                try:
                    self.stop_line.remove()
                except (NotImplementedError, ValueError):
                    pass
            self.stop_line = None

        self.ui.Window3.canvas.draw()

    def process_all(self):
        """Starts video processing in a separate QThread."""
        if not self.VIDEO_LOADED:
            return

        # Get UI values
        ilow, ihigh = (
            self.ui.spinBox_FirstGoodFrame.value(),
            self.ui.spinBox_LastGoodFrame.value(),
        )
        frame_stride = self.ui.spinBox_frame_skips.value()

        # Update metadata
        self.videometa["FIRST_GOOD_FRAME"] = ilow
        self.videometa["LAST_GOOD_FRAME"] = ihigh
        self.videometa.write()

        # ✅ Stop existing thread before starting a new one
        if hasattr(self, "thread") and isinstance(self.thread, QThread):
            if self.thread.isRunning():
                print("⚠️ Stopping existing worker thread...")
                self.worker.stop()  # Ensure worker has a `stop()` method
                self.thread.quit()
                self.thread.wait(5000)  # Wait for thread to exit
                print("✅ Previous worker thread stopped successfully")

        # ✅ Ensure proper cleanup of old worker
        if hasattr(self, "worker") and self.worker is not None:
            self.worker.deleteLater()
            self.worker = None

        # ✅ Create a new Worker Thread safely
        self.worker = ProcessorWorker(
            self.processor,
            self.video,
            self.grab_ui_values(),
            ilow,
            ihigh,
            frame_stride,
            self.ui.lineEdit_filename.text(),
            write_json=True,
            write_video=self.ui.checkBox_writeVideo.isChecked(),
            display_shock=self.ui.checkBox_display_shock.isChecked(),
        )

        self.thread = QThread()
        self.worker.moveToThread(self.thread)
    
        # ✅ Connect signals (only once)
        self.thread.started.connect(self.worker.run)
        self.worker.progress_updated.connect(self.ui.progressBar.setValue)
        self.worker.finished.connect(self.on_processing_complete)
        self.worker.finished.connect(self.thread.quit)
        self.thread.finished.connect(lambda: print("✅ Worker thread fully terminated"))

        # ✅ Start the thread
        self.thread.start()

    def on_processing_complete(self):
        """Handles UI updates when processing is completed."""
        if self.thread:
            print("Stopping thread safely...")
            try:
                self.worker.finished.disconnect()
            except:
                pass
            self.thread.quit()
            self.thread.wait(5000)
            print("✅ Worker thread stopped successfully")

            self.worker.deleteLater()
            self.worker = None
            self.thread.deleteLater()
            self.thread = None

        self.processing_done = True
        self.ui.progressBar.setValue(0)
        self.arcjetcv_message_box("Video Processed", "The video has been processed.")


    def grab_ui_values(self):
        inputdict = {"SEGMENT_METHOD": str(self.ui.comboBox_filterType.currentText())}
        inputdict["HSV_MODEL_RANGE"] = [
            (
                self.ui.minHue.value(),
                self.ui.minSaturation.value(),
                self.ui.minIntensity.value(),
            ),
            (
                self.ui.maxHue.value(),
                self.ui.maxSaturation.value(),
                self.ui.maxIntensity.value(),
            ),
        ]
        inputdict["HSV_SHOCK_RANGE"] = [
            (
                self.ui.minHue_2.value(),
                self.ui.minSaturation_2.value(),
                self.ui.minIntensity_2.value(),
            ),
            (
                self.ui.maxHue_2.value(),
                self.ui.maxSaturation_2.value(),
                self.ui.maxIntensity_2.value(),
            ),
        ]
        inputdict["THRESHOLD"] = self.ui.minIntensity.value()
        return inputdict

    # PLOTTING DATA TAB
    def load_outputs(self):
        try:
            # create fileDialog to select file
            options = QtWidgets.QFileDialog.Options()
            files, _ = QtWidgets.QFileDialog.getOpenFileNames(
                self,
                "Load ouput files",
                "",
                "Output Files (*.json);;All Files (*)",
                options=options,
            )
            self.ui.basebar.setText("Loading %i files" % len(files))

            # Load all files & concatenate
            self.raw_outputs = []
            for fname in files:
                with open(fname, "rb") as file:
                    opl = json.load(file)
                    self.raw_outputs.extend(opl)

            if len(files) > 0:
                fpath, name, ext = splitfn(files[0])
                shortpath = self.shorten_path(fpath, 30)

                # Show summary of loaded data
                summary = "Loaded %i files\n" % len(files)
                summary += "Folder: %s\n" % shortpath
                for fname in files:
                    fpath, name, ext = splitfn(fname)
                    summary += "File: %s\n" % name
                summary += "Total frames: %i\n" % len(self.raw_outputs)

                self.ui.label_data_summary.setText(summary)
                self.ui.basebar.setText("Finished loading files")

        except Exception as e:
            if self.testing:
                print("File loading failed:\n" + str(e))
            else:
                self.arcjetcv_message_box("Warning", "File loading failed:\n" + str(e))

        self.plot_outputs()

    @staticmethod
    def shorten_path(path, max_length):
        if len(path) <= max_length:
            return path  # Return the original path if it is already short enough

        max_visible_length = max_length - 3  # Reserve 3 characters for '...'
        first_part = path[: max_visible_length // 2]
        last_part = path[-(max_visible_length - len(first_part)) :]
        return first_part + "..." + last_part

    def _apply_kalman_filter(self, series):
        if series is None:
            return series

        is_masked = isinstance(series, np.ma.MaskedArray)
        if is_masked:
            data = np.array(series.filled(np.nan), dtype=float)
            mask = np.ma.getmaskarray(series).copy()
        else:
            data = np.array(series, dtype=float)
            mask = np.zeros_like(data, dtype=bool)

        non_finite_mask = ~np.isfinite(data)
        if non_finite_mask.any():
            mask = mask | non_finite_mask
            data[non_finite_mask] = np.nan

        valid_mask = ~mask
        if np.count_nonzero(valid_mask) <= 1:
            return series

        valid_values = data[valid_mask]
        measurement_var = np.nanvar(valid_values)
        if not np.isfinite(measurement_var) or measurement_var <= 1e-12:
            measurement_var = 1.0

        process_var = measurement_var * 0.05
        if not np.isfinite(process_var) or process_var <= 0:
            process_var = 1e-3

        error_cov = measurement_var
        filtered = data.copy()
        valid_indices = np.flatnonzero(valid_mask)
        estimate = data[valid_indices[0]]
        filtered[valid_indices[0]] = estimate

        for idx in valid_indices[1:]:
            error_cov += process_var
            kalman_gain = error_cov / (error_cov + measurement_var)
            estimate = estimate + kalman_gain * (data[idx] - estimate)
            error_cov = (1.0 - kalman_gain) * error_cov
            filtered[idx] = estimate

        if is_masked:
            return np.ma.masked_array(filtered, mask=mask)

        filtered[~valid_mask] = data[~valid_mask]
        return filtered

    def plot_outputs(self):
        self.ui.basebar.setText("Setting up plots...")

        # Reset plotting windows
        self.ui.Window1.canvas.axes.clear()
        self.ui.Window2.canvas.axes.clear()

        self.ax1 = self.ui.Window1.canvas.axes
        self.ax2 = self.ui.Window2.canvas.axes

        # Plotting params
        index, m95, m50, mc, p50, p95, radius = [], [], [], [], [], [], []
        time, sarea, marea, sc, sm, ypos = [], [], [], [], [], []

        units_selected = self.ui.comboBox_units.currentText()
        units_lower = units_selected.lower()
        use_mm_requested = "mm" in units_lower

        pixels_per_mm_default = None
        if self.raw_outputs and "PIXELS_PER_MM" in self.raw_outputs[0]:
            pixels_per_mm_default = self.raw_outputs[0]["PIXELS_PER_MM"]

        default_pixel_length = (
            1 / pixels_per_mm_default
            if pixels_per_mm_default not in (None, 0, float("inf"))
            else None
        )

        use_mm = use_mm_requested and default_pixel_length is not None
        units = "[mm]" if use_mm else ("[px]" if use_mm_requested else units_selected)
        xy_scale_default = default_pixel_length if use_mm else 1.0
        length_unit_label = units.strip()
        if length_unit_label.startswith("[") and length_unit_label.endswith("]"):
            length_unit_label = length_unit_label[1:-1]
        if not length_unit_label:
            length_unit_label = units

        fps = self.ui.doubleSpinBox_fps.value()
        maskn = self.ui.spinBox_mask_frames.value()
        flow_sign = self._flow_axis_sign(self.ui.comboBox_flowDirection.currentText())

        self.ui.basebar.setText("Plotting data...")
        try:
            if len(self.raw_outputs) > 10:
                numframes = len(self.raw_outputs)
                max_traces = 10
                first_color = "red"
                swell_color = "#2e8b57"
                last_color = "#1f77b4"
                mid_color = "#999999"

                model_traces = []
                shock_traces = {}
                swell_trace_index = None

                for loop_idx, i in enumerate(range(0, len(self.raw_outputs), maskn)):
                    # Save frame index, time
                    index.append(self.raw_outputs[i]["INDEX"])
                    time.append(self.raw_outputs[i]["INDEX"] / fps)

                    frame_scale = xy_scale_default
                    frame_pixels_per_mm = self.raw_outputs[i].get(
                        "PIXELS_PER_MM", pixels_per_mm_default
                    )
                    if use_mm and frame_pixels_per_mm not in (None, 0, float("inf")):
                        frame_scale = 1 / frame_pixels_per_mm

                    # Model positions (-95%, -50%, center, 50%, 95% radius)
                    if self.raw_outputs[i]["MODEL"] is not None:
                        xpos = np.asarray(
                            self.raw_outputs[i]["MODEL_INTERP_XPOS"], dtype=float
                        )
                        signed_xpos = flow_sign * xpos
                        center = self.raw_outputs[i]["MODEL_YCENTER"]
                        model_array = np.array(self.raw_outputs[i]["MODEL"])
                        self.raw_outputs[i]["MODEL"] = model_array
                        model_coords = np.array(model_array[:, 0, :], dtype=float)
                        if use_mm:
                            model_coords *= frame_scale
                        model_traces.append(
                            {
                                "loop_idx": loop_idx,
                                "coords": model_coords,
                                "centroid": self.raw_outputs[i].get(
                                    "MODEL_CENTROID_X"
                                ),
                            }
                        )
                        m95.append(signed_xpos[0])
                        m50.append(signed_xpos[1])
                        mc.append(signed_xpos[2])
                        p50.append(signed_xpos[3])
                        p95.append(signed_xpos[4])
                        ypos.append(center)
                        radius.append(self.raw_outputs[i]["MODEL_RADIUS"])
                    else:
                        m95.append(np.nan)
                        m50.append(np.nan)
                        mc.append(np.nan)
                        p50.append(np.nan)
                        p95.append(np.nan)
                        ypos.append(np.nan)
                        radius.append(np.nan)

                    # Shock center x-position
                    if self.raw_outputs[i]["SHOCK"] is not None:
                        sc.append(
                            flow_sign
                            * float(self.raw_outputs[i]["SHOCK_INTERP_XPOS"][0])
                        )
                        shock_array = np.array(self.raw_outputs[i]["SHOCK"])
                        self.raw_outputs[i]["SHOCK"] = shock_array
                        shock_coords = np.array(shock_array[:, 0, :], dtype=float)
                        if use_mm:
                            shock_coords *= frame_scale
                        shock_traces[loop_idx] = shock_coords
                    else:
                        sc.append(np.nan)

                    # Shock and model area
                    if self.raw_outputs[i]["SHOCK"] is not None:
                        sarea.append(self.raw_outputs[i]["SHOCK_AREA"])
                    else:
                        sarea.append(np.nan)

                    if self.raw_outputs[i]["MODEL"] is not None:
                        marea.append(self.raw_outputs[i]["MODEL_AREA"])
                    else:
                        marea.append(np.nan)

                    # Shock-model separation, center
                    if (
                        self.raw_outputs[i]["MODEL"] is not None
                        and self.raw_outputs[i]["SHOCK"] is not None
                    ):
                        sm.append(abs(sc[-1] - mc[-1]))
                    else:
                        sm.append(np.nan)

                if model_traces:
                    # Determine swell index based on minimum centroid value
                    valid_centroids = [
                        (idx, trace["centroid"])
                        for idx, trace in enumerate(model_traces)
                        if trace["centroid"] is not None
                    ]
                    if valid_centroids:
                        swell_trace_index = min(valid_centroids, key=lambda x: x[1])[0]

                    num_model_traces = len(model_traces)
                    stride = max(1, num_model_traces // max_traces)
                    selected_indices = set(range(0, num_model_traces, stride))
                    selected_indices.add(0)
                    selected_indices.add(num_model_traces - 1)
                    if swell_trace_index is not None:
                        selected_indices.add(swell_trace_index)
                    selected_indices = sorted(selected_indices)

                    used_labels = set()

                    def style_for_trace(trace_idx):
                        if trace_idx == 0:
                            return first_color, 2.5, "Initial Profile"
                        if (
                            swell_trace_index is not None
                            and trace_idx == swell_trace_index
                        ):
                            return swell_color, 1.0, "Growth Profile"
                        if trace_idx == num_model_traces - 1:
                            return last_color, 1.0, "Final Profile"
                        return mid_color, 1.0, "Model Profile"

                    for trace_idx in selected_indices:
                        trace = model_traces[trace_idx]
                        coords = trace["coords"]
                        if self.ui.checkBox_kalman_filter.isChecked():
                            model_x = self._apply_kalman_filter(coords[:, 0])
                            model_y = self._apply_kalman_filter(coords[:, 1])
                            coords = np.column_stack(
                                (np.asarray(model_x, dtype=float), np.asarray(model_y, dtype=float))
                            )
                        color, linewidth, label = style_for_trace(trace_idx)
                        display_label = (
                            label if label and label not in used_labels else None
                        )
                        if display_label:
                            used_labels.add(display_label)

                        self.ax1.plot(
                            coords[:, 0],
                            coords[:, 1],
                            color=color,
                            linewidth=linewidth,
                            alpha=0.9 if display_label else 0.5,
                            label=display_label,
                        )

                        if (
                            self.ui.checkBox_display_shock2.isChecked()
                            and trace["loop_idx"] in shock_traces
                        ):
                            shock_coords = shock_traces[trace["loop_idx"]]
                            if self.ui.checkBox_kalman_filter.isChecked():
                                shock_x = self._apply_kalman_filter(shock_coords[:, 0])
                                shock_y = self._apply_kalman_filter(shock_coords[:, 1])
                                shock_coords = np.column_stack(
                                    (
                                        np.asarray(shock_x, dtype=float),
                                        np.asarray(shock_y, dtype=float),
                                    )
                                )
                            shock_label = (
                                "Shock Profile"
                                if "Shock Profile" not in used_labels
                                else None
                            )
                            if shock_label:
                                used_labels.add(shock_label)
                            self.ax1.plot(
                                shock_coords[:, 0],
                                shock_coords[:, 1],
                                color="black",
                                linewidth=0.8,
                                linestyle="--",
                                alpha=0.5,
                                label=shock_label,
                            )

                    self.swell_loop_idx = (
                        model_traces[swell_trace_index]["loop_idx"]
                        if swell_trace_index is not None
                        else None
                    )
                    self.swell_time = (
                        time[self.swell_loop_idx]
                        if self.swell_loop_idx is not None
                        and self.swell_loop_idx < len(time)
                        else None
                    )
                else:
                    self.swell_loop_idx = None
                    self.swell_time = None

                handles, labels = self.ax1.get_legend_handles_labels()
                if labels:
                    legend_map = {}
                    for handle, label in zip(handles, labels):
                        if not label or label in legend_map:
                            continue
                        legend_map[label] = handle

                    legend = self.ax1.legend(
                        list(legend_map.values()),
                        list(legend_map.keys()),
                        loc="best",
                    )
                    legend.get_title().set_fontweight("bold")

                self.ax1.set_xlabel(f"X ({length_unit_label})")
                self.ax1.set_ylabel(f"Y ({length_unit_label})")
                self.ax1.invert_yaxis()
                self.ax1.figure.canvas.draw()

                # Mask outliers
                metrics = [marea, ypos, radius, mc]
                mask = getOutlierMask(metrics)
                self.XT_MASK = mask

                # Infer pixel length from raw outputs
                radius_masked = np.ma.masked_where(mask < 0, radius)
                if default_pixel_length is not None:
                    pixel_length = default_pixel_length
                else:
                    pixel_length = 1  # Default to pixel units if calibration missing
                    if use_mm_requested:
                        self.arcjetcv_message_box(
                            "Warning",
                            "Pixels per mm not available in raw outputs. "
                            "Spatial data will be shown in pixels.",
                        )
                self.pixel_length = pixel_length
                series_scale = pixel_length if use_mm else 1.0
                radius_series = radius_masked * series_scale

                # Plot XT series
                ym95 = np.ma.masked_where(mask < 0, m95) * series_scale
                ym50 = np.ma.masked_where(mask < 0, m50) * series_scale
                ymc = np.ma.masked_where(mask < 0, mc) * series_scale
                yp50 = np.ma.masked_where(mask < 0, p50) * series_scale
                yp95 = np.ma.masked_where(mask < 0, p95) * series_scale
                ysarea = np.ma.masked_where(mask < 0, sarea)
                ymarea = np.ma.masked_where(mask < 0, marea)
                ysc = np.ma.masked_where(mask < 0, sc) * series_scale
                ysm = np.ma.masked_where(mask < 0, sm) * series_scale
                yypos = np.ma.masked_where(mask < 0, ypos) * series_scale

                if self.ui.checkBox_kalman_filter.isChecked():
                    ym95 = self._apply_kalman_filter(ym95)
                    ym50 = self._apply_kalman_filter(ym50)
                    ymc = self._apply_kalman_filter(ymc)
                    yp50 = self._apply_kalman_filter(yp50)
                    yp95 = self._apply_kalman_filter(yp95)
                    ysarea = self._apply_kalman_filter(ysarea)
                    ymarea = self._apply_kalman_filter(ymarea)
                    ysc = self._apply_kalman_filter(ysc)
                    ysm = self._apply_kalman_filter(ysm)
                    yypos = self._apply_kalman_filter(yypos)
                    radius_series = self._apply_kalman_filter(radius_series)

                self.PLOTKEYS = []
                self.max_expansion_info = None
                self.max_recession_info = None

                if self.ui.checkBox_m95_radius.isChecked():
                    self.ax2.plot(
                        time,
                        ym95,
                        color="magenta",
                        linewidth=0,
                        marker="o",
                        markersize=4,
                        label="Model -95%R",
                    )
                    self.PLOTKEYS.append("MODEL_-0.95R " + units)

                if self.ui.checkBox_m50_radius.isChecked():
                    self.ax2.plot(
                        time,
                        ym50,
                        color="blue",
                        linewidth=0,
                        marker="o",
                        markersize=4,
                        label="Model -50%R",
                    )
                    self.PLOTKEYS.append("MODEL_-0.50R " + units)

                if self.ui.checkBox_model_center.isChecked():
                    time_arr = np.asarray(time, dtype=float)
                    center_mask = ~np.ma.getmaskarray(ymc)
                    time_valid = time_arr[center_mask]
                    center_valid = np.asarray(ymc)[center_mask]

                    if center_valid.size > 0:
                        center_zeroed = center_valid - center_valid[0]
                        exp_idx = int(np.argmin(center_zeroed))
                        rec_idx = int(np.argmax(center_zeroed))
                        self.max_expansion_info = (
                            float(center_zeroed[exp_idx]),
                            float(time_valid[exp_idx]),
                        )
                        self.max_recession_info = (
                            float(center_zeroed[rec_idx]),
                            float(time_valid[rec_idx]),
                        )
                        swell_idx = int(np.argmin(center_zeroed))
                        swell_time_value = time_valid[swell_idx]

                        self.ax2.axvline(
                            swell_time_value,
                            ymin=0.0,
                            ymax=1.0,
                            linestyle="--",
                            color="black",
                            linewidth=1,
                            label=f"Recess Start: {swell_time_value:.2f} [s]",
                        )

                        if swell_idx >= 0:
                            self.ax2.plot(
                                time_valid[: swell_idx + 1],
                                center_zeroed[: swell_idx + 1],
                                color=swell_color,
                                linewidth=0,
                                marker="o",
                                markersize=5,
                                label="Growth",
                            )

                        if swell_idx + 1 < center_zeroed.size:
                            recession_time = time_valid[swell_idx + 1 :]
                            recession_values = center_zeroed[swell_idx + 1 :]
                            self.ax2.plot(
                                recession_time,
                                recession_values,
                                color=last_color,
                                linewidth=0,
                                marker="o",
                                markersize=5,
                                label="Recess",
                            )
                            if recession_time.size >= 2:
                                coeffs = np.polyfit(
                                    recession_time, recession_values, 1
                                )
                                fit_line = np.polyval(coeffs, recession_time)
                                self.ax2.plot(
                                    recession_time,
                                    fit_line,
                                    color="red",
                                    linewidth=1,
                                    label=f"Fit: y = {coeffs[0]:.2f}x+{coeffs[1]:.2f}",
                                )

                        valid_indices = np.flatnonzero(center_mask)
                        if swell_idx < valid_indices.size:
                            self.swell_loop_idx = int(valid_indices[swell_idx])
                            self.swell_time = time[self.swell_loop_idx]

                        # Keep dynamic plot limits; fixed ticks can force misleading ranges.
                        y_min = float(np.min(center_zeroed))
                        y_max = float(np.max(center_zeroed))
                        if np.isfinite(y_min) and np.isfinite(y_max):
                            lower_data = min(y_min, 0.0)
                            upper_data = max(y_max, 0.0)
                            span = upper_data - lower_data
                            pad = max(0.25, 0.1 * span) if span > 0 else 0.5
                            self.ax2.set_ylim(lower_data - pad, upper_data + pad)
                        self.ax2.locator_params(axis="x", nbins=6)
                        self.ax2.locator_params(axis="y", nbins=6)

                    self.PLOTKEYS.append("MODEL_CENTER " + units)

                if self.ui.checkBox_50_radius.isChecked():
                    self.ax2.plot(
                        time,
                        yp50,
                        color="cyan",
                        linewidth=0,
                        marker="o",
                        markersize=4,
                        label="Model +50%R",
                    )
                    self.PLOTKEYS.append("MODEL_0.50R " + units)

                if self.ui.checkBox_95_radius.isChecked():
                    self.ax2.plot(
                        time,
                        yp95,
                        color="red",
                        linewidth=0,
                        marker="o",
                        markersize=4,
                        label="Model +95%R",
                    )
                    self.PLOTKEYS.append("MODEL_0.95R " + units)

                if self.ui.checkBox_shock_area.isChecked():
                    self.ax2.plot(
                        time,
                        ysarea,
                        color="gold",
                        linewidth=0,
                        marker="o",
                        markersize=4,
                        label="Shock area (px)",
                    )
                    self.PLOTKEYS.append("SHOCK_AREA [px]")

                if self.ui.checkBox_model_rad.isChecked():
                    self.ax2.plot(
                        time,
                        radius_series,
                        color="navy",
                        linewidth=0,
                        marker="o",
                        markersize=4,
                        label=f"Model radius ({length_unit_label})",
                    )
                    self.PLOTKEYS.append("MODEL_RADIUS " + units)

                if self.ui.checkBox_shock_center.isChecked():
                    self.ax2.plot(
                        time,
                        ysc,
                        color="darkcyan",
                        linewidth=0,
                        marker="o",
                        markersize=4,
                        label="Shock center",
                    )
                    self.PLOTKEYS.append("SHOCK_CENTER " + units)

                if self.ui.checkBox_shockmodel.isChecked():
                    self.ax2.plot(
                        time,
                        ysm,
                        color="darkred",
                        linewidth=0,
                        marker="o",
                        markersize=4,
                        linestyle="--",
                        label="Shock-model distance",
                    )
                    self.PLOTKEYS.append("SHOCK_TO_MODEL " + units)

                if self.ui.checkBox_ypos.isChecked():
                    self.ax2.plot(
                        time,
                        yypos,
                        color="black",
                        linewidth=0,
                        marker="o",
                        markersize=4,
                        label="Axis position",
                    )
                    self.PLOTKEYS.append("MODEL_YPOS " + units)

                self.ax2.set_xlabel("Time (s)")
                self.ax2.set_ylabel(units)

                legend = self.ax2.legend()
                legend.set_draggable(True)
                self.ax2.figure.canvas.draw()

                # Save to dictionary data structure
                output_dict = {"TIME [s]": time}
                self.length_units = [ym95, ym50, ymc, yp50, yp95, ysc, ysm, yypos]
                self.length_labels = [
                    "MODEL_-0.95R",
                    "MODEL_-0.50R",
                    "MODEL_CENTER",
                    "MODEL_0.50R",
                    "MODEL_0.95R",
                    "SHOCK_CENTER",
                    "SHOCK_TO_MODEL",
                    "MODEL_YPOS",
                ]
                for k in range(0, len(self.length_units)):
                    output_dict[self.length_labels[k] + " " + units] = (
                        self.length_units[k]
                    )

                self.px_units = [ymarea, ysarea, radius_series]
                self.px_labels = [
                    "MODEL_AREA [px]",
                    "SHOCK_AREA [px]",
                    "MODEL_RADIUS " + units,
                ]
                for k in range(0, len(self.px_units)):
                    output_dict[self.px_labels[k]] = self.px_units[k]

                output_dict["CONFIG"] = [
                    "UNITS: %s" % units,
                    "FPS: %.2f" % fps,
                    "MASK_NFRAMES: %i" % maskn,
                ]
                self.time_series = output_dict.copy()

                # Update ui metrics
                self.ui.doubleSpinBox_fit_start_time.setMinimum(min(time))
                self.ui.doubleSpinBox_fit_start_time.setMaximum(max(time))
                self.ui.doubleSpinBox_fit_last_time.setMinimum(min(time))
                self.ui.doubleSpinBox_fit_last_time.setMaximum(max(time))
                self.ui.doubleSpinBox_fit_start_time.setValue(min(time))
                self.ui.doubleSpinBox_fit_last_time.setValue(max(time))

                # Update data summary with pixel length and displacement statistics
                summary = self.ui.label_data_summary.text().strip()
                lines = summary.split("\n") if summary else []
                px_per_mm = (
                    1 / pixel_length if pixel_length != 0 else float("inf")
                )  # Avoid division by zero

                def upsert_line(prefix, text):
                    for idx, line in enumerate(lines):
                        if line.startswith(prefix):
                            lines[idx] = text
                            return
                    lines.append(text)

                def remove_line(prefix):
                    lines[:] = [line for line in lines if not line.startswith(prefix)]

                upsert_line(
                    "Pixel length",
                    "Pixel length %s: %.4f mm/px | %.4f px/mm"
                    % (
                        self.ui.comboBox_units.currentText(),
                        pixel_length,
                        px_per_mm,
                    ),
                )

                if self.max_expansion_info is not None:
                    exp_val, exp_time = self.max_expansion_info
                    upsert_line(
                        "Max expansion",
                        "Max expansion: %+0.3f %s at %.2f s"
                        % (exp_val, length_unit_label, exp_time),
                    )
                else:
                    remove_line("Max expansion")

                if self.max_recession_info is not None:
                    rec_val, rec_time = self.max_recession_info
                    upsert_line(
                        "Max recession",
                        "Max recession: %+0.3f %s at %.2f s"
                        % (rec_val, length_unit_label, rec_time),
                    )
                else:
                    remove_line("Max recession")

                self.ui.label_data_summary.setText("\n".join(lines))
                self.ui.basebar.setText("Finished plotting data")
            else:
                self.ui.basebar.setText("Not enough data to plot")
                if self.testing:
                    print("Not enough data to plot")
                else:
                    self.arcjetcv_message_box(
                        "Warning",
                        "Not enough data to plot: only %i points"
                        % len(self.raw_outputs),
                    )

        except Exception as e:
            self.ui.basebar.setText("Plotting failed !")
            if self.testing:
                print("Warning", "Plotting failed :\n" + str(e))
            else:
                self.arcjetcv_message_box("Warning", "Plotting failed:\n" + str(e))

        self.ui.Window1.repaint()
        self.ui.Window2.repaint()
        self.fit_data()

    def save_plots(self, name):
        try:
            # Extract necessary data from self.time_series
            time = self.time_series["TIME [s]"]
            units = self.ui.comboBox_units.currentText()

            # Plot each dataset in a separate graph
            for data, label in zip(self.length_units[5:], self.length_labels[5:]):
                fig, ax = plt.subplots()
                ax.scatter(time, data, label=label)
                ax.set_xlabel("Time (s)")
                ax.set_ylabel(f"{label} ({units})")
                ax.legend()
                filename = f"_{label.replace(' ', '_').lower()}.png"
                fig.savefig(str(name + filename))
                plt.close(fig)

                # Plot each dataset in a separate graph
            for data, label in zip(self.px_units, self.px_labels):
                fig, ax = plt.subplots()
                ax.scatter(time, data, label=label)
                ax.set_xlabel("Time (s)")
                ax.set_ylabel(f"{label} ({units})")
                ax.legend()
                filename = f"_{label.replace(' ', '_').lower()}.png"
                fig.savefig(str(name + filename))
                plt.close(fig)

            # Plot "MODEL_-0.95R", "MODEL_-0.50R", "MODEL_CENTER", "MODEL_0.50R", "MODEL_0.95R" in one graph
            fig, ax_combined = plt.subplots()
            for i, data in enumerate(self.length_units[:5]):
                label = f"Model {['-0.95R', '-0.50R', 'Center', '0.50R', '0.95'][i]}"
                ax_combined.scatter(time, data, label=label)
            ax_combined.set_xlabel("Time (s)")
            ax_combined.set_ylabel(f"Model Positions ({units})")
            ax_combined.legend()
            fig.savefig(name + "_radius.png")
            plt.close(fig)

        except Exception as e:
            self.ui.basebar.setText("Plot saving failed !")
            if self.testing:
                print("Warning", "Plot saving failed:\n" + str(e))
            else:
                self.arcjetcv_message_box(
                    "Warning", "Plot saving failed:\n" + str(e)
                )

    def export_to_csv(self):
        if self.time_series is not None:
            try:
                dialog = QtWidgets.QFileDialog()
                pathmask = dialog.getExistingDirectory(None, "Export CSV/plots", "")
                os.makedirs(pathmask, exist_ok=True)
                folder_path = os.path.abspath(pathmask)
                folder_name = os.path.basename(pathmask)

                self.ax1.figure.savefig(
                    os.path.join(folder_path, folder_name + "_xy.png")
                )
                self.ax2.figure.savefig(
                    os.path.join(folder_path, folder_name + "_xt.png")
                )
                self.save_plots(os.path.join(folder_path, folder_name))

                # Convert time series into dataframe
                df = pd.DataFrame(
                    dict([(k, pd.Series(v)) for k, v in self.time_series.items()])
                )

                # Convert fits into dataframe
                if self.fit_dict is not None:
                    param_dict = {"Pixel length": self.pixel_length}
                    # Create pandas Series from fitparam dictionary
                    keys, vals, errs = [], [], []
                    for key, val in self.fit_dict.items():
                        if "LINEAR_FIT" in key:
                            keys.append(key + "_slope")
                            keys.append(key + "_intercept")
                            vals.append(val[0][0])
                            vals.append(val[0][1])
                            errs.append(val[1][0])
                            errs.append(val[1][1])

                    # Create pandas Series from fitparam dictionary
                    keys_series = pd.Series(keys)
                    values_series = pd.Series(vals)
                    err_series = pd.Series(errs)
                    param_dict.update(self.fit_dict)

                    # Assign the Series as new columns to the DataFrame
                    df = df.assign(
                        Keys=keys_series, Values=values_series, Error=err_series
                    )

                    csv_file_path = os.path.join(folder_path, f"{folder_name}.csv")
                    df.to_csv(csv_file_path, index=False)

            except Exception as e:
                self.ui.basebar.setText("CSV export failed !")
                if self.testing:
                    print("Warning", "CSV export failed:\n" + str(e))
                else:
                    self.arcjetcv_message_box(
                        "Warning", "CSV export failed:\n" + str(e)
                    )

    def fit_data(self):
        if self.time_series is not None:

            try:
                # Retrieve user inputs for time/units
                units = self.ui.comboBox_units.currentText()
                time = np.array(self.time_series["TIME [s]"])
                mask = np.array(self.XT_MASK)
                t0 = self.ui.doubleSpinBox_fit_start_time.value()
                t1 = self.ui.doubleSpinBox_fit_last_time.value()

                # identify relevant indicies
                # Include interval endpoints to avoid dropping valid edge samples.
                inds = (time >= t0) * (time <= t1) * (mask > 0)

                # Initialize data structure
                fit_dict = {}

                # Get list of keys to loop through
                keys = list(self.time_series.keys())
                keys.remove("TIME [s]")
                keys.remove("CONFIG")

                # Linear fits
                if self.ui.comboBox_fit_type.currentText() == "linear":
                    longstring = ""
                    for key in keys:
                        t = time[inds]
                        y = self.time_series[key][inds]
                        if len(t) > 2 and len(y) > 2:
                            fitp, cov = np.ma.polyfit(t, y, 1, cov=True)
                            err = np.sqrt(np.diag(cov))
                            m, b = fitp[0], fitp[1]
                            fit_dict[key + "_LINEAR_FIT"] = (fitp, err)
                            if key in self.PLOTKEYS:
                                longstring += (
                                    key
                                    + "\tMin = %f\t Max = %f\t Delta = %f\n"
                                    % (y.min(), y.max(), y.max() - y.min())
                                )
                                longstring += "\tLINEAR FIT: y = mx+b \t"
                                longstring += (
                                    "m = %f+-%f %s" % (m, err[0], units + "/s") + "\t"
                                )
                                longstring += (
                                    "b = %f+-%f %s" % (b, err[1], units) + "\n\n"
                                )
                        else:
                            longstring = "not enough points to fit data"

                    self.ui.textBrowser.setText(longstring)
                    self.ui.textBrowser.repaint()

                self.fit_dict = fit_dict

            except Exception as e:
                self.ui.basebar.setText("Fitting failed !")
                if self.testing:
                    print("Warning", "Fitting failed :\n" + str(e))
                else:
                    self.arcjetcv_message_box(
                        "Warning", "Fitting failed :\n" + str(e)
                    )

    def arcjetcv_message_box(self, title, message):
        msg_box = QMessageBox(self)
        msg_box.setIconPixmap(QPixmap(self.logo_path))
        msg_box.setWindowTitle(title)
        msg_box.setText(message)
        msg_box.setStandardButtons(QMessageBox.Ok)
        msg_box.exec()



from PySide6 import QtCore, QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas, NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure


class TrackLabel(QtWidgets.QLabel):
    newCursorValue = QtCore.Signal(list)

    def mouseMoveEvent(self,event):
        self.newCursorValue.emit([event.x(), event.y(),self.width(), self.height()])


class MplCanvas(FigureCanvas):
    """ Convenience class to embed matplotlib canvas

    Args:
        FigureCanvas (matplotlib object): matplotlib canvas object
    """
    clicked = QtCore.Signal(float, float)

    def __init__(self, parent=None, width=5, height=4, dpi=100):
        self.figure = Figure(figsize=(width, height), dpi=dpi)
        self.figure.set_layout_engine('tight')
        self.axes = self.figure.add_subplot(111)
        super(MplCanvas, self).__init__(self.figure)

        # Connect the mouse click event to the callback function
        self.mpl_connect("button_press_event", self.on_mouse_click)

    def on_mouse_click(self, event):
        if event.button == 1:  # Left mouse button
            x, y = event.xdata, event.ydata
            if x is not None and y is not None:
                self.clicked.emit(x, y)


class MatplotlibWidget(QtWidgets.QWidget):
    """Plotting widget using matplotlib, embedding into Qt

    Args:
        MplCanvas (FigureCanvas): matplotlib standard canvas
    """
    
    def __init__(self,*args):
        super().__init__(*args)
        layout = QtWidgets.QVBoxLayout(self)
        self.canvas = MplCanvas(self)
        toolbar = NavigationToolbar(self.canvas, self)
        layout.addWidget(toolbar)
        layout.addWidget(self.canvas)
        self.setLayout(layout)


class OverlayLabel(QtWidgets.QLabel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowFlags(QtCore.Qt.FramelessWindowHint | QtCore.Qt.Tool)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)

    def showEvent(self, event):
        super().showEvent(event)
        self.resize(self.parent().size())



import os
import json
import cv2 as cv
import numpy as np
import threading
from arcjetCV.utils.utils import splitfn
try:
    from arcjetCV.segmentation.time.time_segmentation import (
        time_segmentation,
        extract_interest,
    )
    _time_segmentation_error = None
except ModuleNotFoundError as exc:
    time_segmentation = None
    extract_interest = None
    _time_segmentation_error = exc


class Video(object):
    """
    Convenience wrapper for opencv video capture.

    This class provides a simple interface for working with video files using OpenCV. It encapsulates functionality for reading frames from a video file, setting the current frame, and writing processed frames to a new video file.

    Attributes:
        fpath (str): Path to the video file.
        name (str): Name of the video file.
        folder (str): Directory containing the video file.
        ext (str): Extension of the video file.
        cap: OpenCV VideoCapture object for reading video frames.
        nframes (int): Total number of frames in the video.
        fps (float): Frames per second of the video.
        shape (tuple): Shape of the video frames (height, width, channels).
        last_frame: Last frame read from the video.
        writer: OpenCV VideoWriter object for writing processed frames.
        _lock: Threading lock for thread-safe access to the video capture object.

    Example:
    ```python
    video = Video('input_video.mp4')
    print(video)
    frame = video.get_frame(0)
    ```
    """

    def __init__(self, path):
        """
        Initializes the Video object.

        :param path: path to the video file
        """
        ### path variables
        self.fpath = path
        folder, name, ext = splitfn(path)
        self.name = name
        self.folder = folder
        self.ext = ext
        self._lock = threading.Lock()

        ### opencv video file object
        self.cap = cv.VideoCapture(self.fpath)
        self.nframes = int(self.cap.get(cv.CAP_PROP_FRAME_COUNT))
        self.fps = self.cap.get(cv.CAP_PROP_FPS)
        ret, frame = self.cap.read()
        self.shape = np.shape(frame)
        self.h, self.w, self.chan = self.shape
        self.last_frame = frame
        print(f"Loaded {self.fpath} with {self.nframes} frames and shape {self.shape}")

        if self.chan not in [1, 3]:
            raise IndexError(
                "ERROR: number of channels of input video (%i) is not 1 or 3!"
                % self.chan
            )

        ### video output
        self.writer = None
        self.output_path = os.path.join(self.folder, "video_out_" + self.name + ".m4v")

    def __str__(self):
        """
        Returns a string representation of the Video object.
        """
        return "Video: {}, shape={}, nframes={}".format(
            self.fpath,
            self.shape,
            self.nframes,
        )

    def get_next_frame(self):
        """
        Retrieves the next frame from the video.

        :returns: next frame
        """
        with self._lock:
            _, self.last_frame = self.cap.read()
            return self.last_frame

    def set_frame(self, index):
        """
        Sets the frame at the specified index.

        :param index: index of the frame to set
        """
        with self._lock:
            self.cap.set(cv.CAP_PROP_POS_FRAMES, index)
            return

    def get_frame(self, index):
        """
        Retrieves the frame at the specified index.

        :param index: index of the frame to retrieve
        :returns: RGB frame at the specified index
        """
        with self._lock:
            self.cap.set(cv.CAP_PROP_POS_FRAMES, index)
            _, self.last_frame = self.cap.read()
            return cv.cvtColor(self.last_frame, cv.COLOR_BGR2RGB)

    def close(self):
        """
        Closes the video capture.
        """
        if self.writer is not None:
            self.writer.release()
        self.cap.release()

    def get_writer(self, video_output_name):
        """
        Initializes the video writer.
        """
        video_output_path = os.path.join(self.folder, video_output_name)
        vid_cod = cv.VideoWriter_fourcc("m", "p", "4", "v")
        print(f"Writing {video_output_path}")
        self.writer = cv.VideoWriter(
            video_output_path, vid_cod, self.fps, (self.shape[1], self.shape[0])
        )

    def close_writer(self):
        """
        Closes the video writer.
        """
        self.writer.release()


class VideoMeta(dict):
    """
    Subclass of dictionary designed to save/load video metadata in JSON format.

    This class extends the dictionary class to provide functionality for saving and loading video metadata in JSON format. It also includes methods for resetting frame crop parameters and setting frame crop parameters.

    Attributes:
        folder (str): Directory containing the video metadata file.
        name (str): Name of the video metadata file.
        ext (str): Extension of the video metadata file.
        path (str): Path to the video metadata file.

    Example:
    ```python
    video = Video('input_video.mp4') # load video
    video_meta = VideoMeta(video, 'metadata.json') # create/load metadata obj
    video_meta.write() # write metadata to file
    ```
    """

    def __init__(self, video, path, progress_callback=None):
        """
        Initializes the VideoMeta object.

        :param video: Video object
        :param path: path to the video file
        """
        super(VideoMeta, self).__init__()
        folder, name, ext = splitfn(path)
        self.folder = folder
        self.name = name
        self.ext = ext
        self.path = path
        self._progress_callback = progress_callback

        def _emit_progress(value, message):
            if self._progress_callback is not None:
                self._progress_callback(int(max(0, min(100, value))), message)

        ### Meta Parameters for each video
        self["WIDTH"] = None
        self["HEIGHT"] = None
        self["CHANNELS"] = None
        self["NFRAMES"] = None
        self["FIRST_GOOD_FRAME"] = None
        self["LAST_GOOD_FRAME"] = None
        self["MODELPERCENT"] = None
        self["FLOW_DIRECTION"] = None
        self["PREAMBLE_RANGE"] = None
        self["STING_VISIBLE_RANGES"] = None
        self["CROP_YMIN"] = None
        self["CROP_YMAX"] = None
        self["CROP_XMIN"] = None
        self["CROP_XMAX"] = None
        self["NOTES"] = None
        self["BRIGHTNESS"] = None

        if os.path.exists(path):
            _emit_progress(15, "Loading metadata...")
            self.load(path)
            _emit_progress(100, "Metadata loaded.")
        else:
            _emit_progress(5, "Initializing metadata...")
            self["WIDTH"] = video.w
            self["HEIGHT"] = video.h
            self["CHANNELS"] = video.chan
            self["NFRAMES"] = video.nframes

            if time_segmentation is not None and extract_interest is not None:
                try:  # Infer meta parameters
                    print("Inferring first and last frames ... ", end="")
                    _emit_progress(10, "Inferring first/last valid frames...")

                    def _time_seg_progress(pct, msg):
                        mapped = 10 + int(35 * (pct / 100.0))
                        _emit_progress(mapped, msg)

                    _, out = time_segmentation(video, progress_callback=_time_seg_progress)
                    start, end = extract_interest(out)
                    print("Done")
                    self["FIRST_GOOD_FRAME"] = max(
                        round(start[0] * video.nframes / 500), int(video.nframes * 0.1)
                    )
                    self["LAST_GOOD_FRAME"] = min(
                        round(video.nframes * end[-1] / 500), int(video.nframes)
                    )
                except Exception as exc:
                    print(f"Time segmentation failed: {exc}")
                    self["FIRST_GOOD_FRAME"] = 0
                    self["LAST_GOOD_FRAME"] = video.nframes
            else:
                if _time_segmentation_error is not None:
                    print(
                        "Time segmentation unavailable "
                        f"({type(_time_segmentation_error).__name__}: "
                        f"{_time_segmentation_error})."
                    )
                self["FIRST_GOOD_FRAME"] = 0
                self["LAST_GOOD_FRAME"] = video.nframes

            # initial crop
            self.reset_frame_crop()

            print("Calculating brightness ... ", end="")
            _emit_progress(45, "Calculating brightness trace...")
            self["BRIGHTNESS"] = []
            video.set_frame(0)
            progress_stride = max(1, video.nframes // 100)
            for idx in range(video.nframes):
                ret, frame = video.cap.read()
                if not ret:
                    break
                gray_frame = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
                brightness = np.mean(gray_frame)
                self["BRIGHTNESS"].append(round(brightness, 2))
                if idx % progress_stride == 0:
                    frame_progress = idx / max(1, video.nframes - 1)
                    _emit_progress(
                        45 + int(50 * frame_progress),
                        "Calculating brightness trace...",
                    )
            print("Done")
            _emit_progress(98, "Writing metadata...")
            self.write()
            _emit_progress(100, "Video metadata ready.")

    def write(self):
        """
        Writes the metadata to a JSON file.
        """
        print(f"Writing {self.path} file ... ", end="")
        fout = open(self.path, "w+")
        json.dump(self, fout)
        fout.close()
        print("Done")

    def load(self, path):
        """
        Loads metadata from a JSON file.

        :param path: path to the JSON file
        """
        fin = open(path, "r")
        dload = json.load(fin)
        self.update(dload)
        fin.close()
        print(f"Loaded {path}")

        ### Ensure path vars are correct
        folder, name, ext = splitfn(path)
        self.folder = folder
        self.name = name
        self.ext = ext
        self.path = path

    def reset_frame_crop(self):
        """
        Resets frame crop parameters to defaults.
        """
        self["CROP_YMIN"] = int(self["HEIGHT"] * 0.10)
        self["CROP_YMAX"] = int(self["HEIGHT"] * 0.90)
        self["CROP_XMIN"] = int(self["WIDTH"] * 0.10)
        self["CROP_XMAX"] = int(self["WIDTH"] * 0.90)

    def set_frame_crop(self, ymin, ymax, xmin, xmax):
        """
        Sets frame crop parameters.

        :param ymin: minimum y coordinate
        :param ymax: maximum y coordinate
        :param xmin: minimum x coordinate
        :param xmax: maximum x coordinate
        """
        self["CROP_YMIN"] = ymin
        self["CROP_YMAX"] = ymax
        self["CROP_XMIN"] = xmin
        self["CROP_XMAX"] = xmax

    def crop_range(self):
        """
        Returns the crop range.
        """
        return [
            [self["CROP_YMIN"], self["CROP_YMAX"]],
            [self["CROP_XMIN"], self["CROP_XMAX"]],
        ]



import numpy as np
import cv2 as cv


def filter_hsv_ranges(hsv: np.ndarray, ranges) -> np.ndarray:
    """
    Filter an HSV image using multiple hue-saturation-value (HSV) ranges and return the combined mask.

    The function applies multiple HSV range filters to the input image and combines the results
    using a bitwise OR operation to create a single mask that captures all the specified ranges.

    Args:
        hsv (np.ndarray): Input HSV image.
        ranges (tuple of two lists of lists): A tuple containing two lists.
            The first list contains the lower bounds of multiple HSV ranges,
            and the second list contains the upper bounds for those same ranges.
            Each bound in the lists is itself a list or tuple of three values (H, S, V).
            The number of range pairs (lower and upper bounds) can vary.

    Returns:
        np.ndarray: A combined binary mask where pixels falling within any of the provided
            HSV ranges are set to 255 (white), and all others are set to 0 (black).

    Examples:
        >>> filter_hsv_ranges(hsv_image, ([ [0, 50, 50], [20, 100, 100] ], [ [10, 255, 255], [30, 150, 150] ]))
        # Returns a mask where only the pixels in two ranges are white:
        # 1) 0<H<10, 50<S<255, 50<V<255
        # 2) 20<H<30, 100<S<150, 100<V<150

    """

    # Initialize a black mask with the same dimensions as the input image
    maskHSV = np.zeros((hsv.shape[0], hsv.shape[1]), dtype=np.uint8)

    # Iterate over each HSV range and apply the filtering
    for i in range(len(ranges[0])):
        # Convert lower and upper bounds to numpy arrays
        lower_bound = np.array(ranges[0][i], dtype=np.uint8)
        upper_bound = np.array(ranges[1][i], dtype=np.uint8)

        mask = cv.inRange(hsv, lower_bound, upper_bound)

        # Combine the current mask with the accumulated mask using bitwise OR
        maskHSV = cv.bitwise_or(mask, maskHSV)

    return maskHSV


def getEdgeFromContour(c, flow_direction, offset=None):
    """
    Find the leading edge of a contour for a given flow direction

    :param c: opencv contour, shape(n,1,n)
    :param flow_direction: 'left', 'right', 'up', or 'down'
    :returns: frontedge, contour of front edge
    """
    ### contours are oriented counterclockwise from top left

    if flow_direction == "right":
        ymin_ind = c[:, 0, 1].argmin()
        ymax_ind = c[:, 0, 1].argmax()
        frontedge = c[ymin_ind:ymax_ind, :, :]
    elif flow_direction == "left":
        ymax_ind = c[:, 0, 1].argmax()
        frontedge = c[ymax_ind:, :, :]
    elif flow_direction == "up":
        xmin_ind = c[:, 0, 0].argmin()
        xmax_ind = c[:, 0, 0].argmax()
        frontedge = c[xmin_ind:xmax_ind:, :, :]
    else:
        xmin_ind = c[:, 0, 0].argmin()
        xmax_ind = c[:, 0, 0].argmax()
        frontedge = np.append(c[xmax_ind:, :, :], c[0:xmin_ind, :, :], axis=0)

    frontedge[:, 0, 1] += offset[0]
    frontedge[:, 0, 0] += offset[1]

    return frontedge


def getPoints(c, flow_direction="right", r=[-0.95, -0.50, 0, 0.50, 0.95], prefix="MODEL"):
    """
    Given an OpenCV contour, this function returns interpolated points at specified
    relative vertical positions to the center of the contour.

    Assumptions:
    1. Only the leading edge is passed into the function (not a closed contour).
    2. The contour is dense, meaning every vertical pixel is occupied.

    Parameters:
    :param c: An OpenCV contour of shape (n, 1, n).
    :param flow_direction (str): Indicates the direction of the flow. Acceptable values
                                 are "right", "left", "up", or "down". Default is "right".
    :param r (list): A list of interpolation points relative to the contour's radius.
                     Default is [-0.95, -0.50, 0, 0.50, 0.95].
    :param prefix (str): A prefix string for keys in the output dictionary. Default is "MODEL".

    Returns:
    :return: A dictionary containing:
        - [prefix]+'_R': List of relative interpolation points.
        - [prefix]+'_YCENTER': Y-coordinate of the center of the contour.
        - [prefix]+'_YLOW': Minimum Y-coordinate of the contour.
        - [prefix]+'_CROP_YMAX': Maximum Y-coordinate of the contour.
        - [prefix]+'_RADIUS': Radius (half of the height) of the contour.
        - [prefix]+'_INTERP_XPOS': Interpolated X-coordinates corresponding to each point in 'r'.
    """

    if flow_direction in ["right", "left"]:
        ### Extract min/max ypos
        low_ind = c[:, 0, 1].argmin()
        ymin = c[low_ind, 0, 1]
        high_ind = c[:, 0, 1].argmax()
        ymax = c[high_ind, 0, 1]
        center = int((ymin + ymax) / 2)
        radius = int((ymax - ymin) / 2)

        ### Setup output dictionary
        output = {
            prefix + "_R": r,
            prefix + "_YCENTER": center,
            prefix + "_YLOW": ymin,
            prefix + "_CROP_YMAX": ymax,
            prefix + "_RADIUS": radius,
        }

        # Interpolate corresponding horizontal positions
        xpos = np.zeros(len(r))
        ypos = [int(y * radius + center) for y in r]

        li = min(low_ind, high_ind)
        hi = max(low_ind, high_ind)
        inds = np.arange(li, hi)
        for ind in inds:
            for j in range(0, len(ypos)):
                if abs(c[ind, 0, 1] - ypos[j]) < 3:
                    xpos[j] = c[ind, 0, 0]
        output[prefix + "_INTERP_XPOS"] = xpos
    else:
        ### Extract min/max ypos
        low_ind = c[:, 0, 0].argmin()
        xmin = c[low_ind, 0, 0]
        high_ind = c[:, 0, 0].argmax()
        xmax = c[high_ind, 0, 0]
        center = int((xmin + xmax) / 2)
        radius = int((xmax - xmin) / 2)

        ### Setup output dictionary
        output = {
            prefix + "_R": r,
            prefix + "_YCENTER": center,
            prefix + "_YLOW": xmin,
            prefix + "_CROP_YMAX": xmax,
            prefix + "_RADIUS": radius,
        }

        # Interpolate corresponding horizontal positions
        ypos = np.zeros(len(r))
        xpos = [int(y * radius + center) for y in r]

        li = min(low_ind, high_ind)
        hi = max(low_ind, high_ind)
        inds = np.arange(li, hi)
        for ind in inds:
            for j in range(0, len(xpos)):
                if abs(c[ind, 0, 0] - xpos[j]) < 3:
                    ypos[j] = c[ind, 0, 1]
        output[prefix + "_INTERP_XPOS"] = ypos

    return output.copy()


def contoursAutoHSV(orig, log=None, flags={"UNDEREXPOSED": False}):
    """
    Find contours using default union of multiple HSV ranges.
    Uses the BGR-HSV transformation to increase contrast.

    :param orig: opencv 8bit BGR image
    :param flags: dictionary with flags
    :param log: log object
    :returns: model contour, shock contour, flags
    """

    img = cv.cvtColor(orig, cv.COLOR_BGR2HSV)

    ### HSV pixel ranges for models taken from sample frames
    model_ranges = np.array([[(0, 0, 208), (155, 0, 155), (13, 20, 101), (0, 190, 100), (12, 150, 130)],
                             [(180, 70, 255), (165, 125, 255), (33, 165, 255), (13, 245, 160), (25, 200, 250)]])
    dim_model = np.array([[(7, 0, 8)], [(20, 185, 101)]])

    ### HSV pixel ranges for shocks taken from sample frames
    shock_ranges = np.array([[(125, 78, 115)], [(145, 190, 230)]])
    dim_shocks = np.array([[(125, 100, 35), (140, 30, 20), (118, 135, 30)], 
                           [(165, 165, 150), (156, 90, 220), (128, 194, 125)]])

    # Append additional ranges for underexposed images
    if flags["UNDEREXPOSED"]:
        model_ranges = np.hstack((model_ranges, dim_model))

    # Apply shock filter and extract shock contour
    shockfilter = filter_hsv_ranges(img, shock_ranges)
    if shockfilter.sum() < 500:
        flags["DIM_SHOCK"] = True
        shock_ranges = np.hstack((shock_ranges, dim_shocks))
        shockfilter = filter_hsv_ranges(img, shock_ranges)
    shockcontours, _ = cv.findContours(shockfilter, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)

    # find the biggest shock contour (shockC) by area
    if len(shockcontours) == 0:
        shockC = None
        flags["SHOCK_CONTOUR_FAILED"] = True
        if log is not None:
            log.write("no shock contours found")
    else:
        shockC = max(shockcontours, key=cv.contourArea)

    # Apply model filter and extract model contour
    modelfilter = filter_hsv_ranges(img, model_ranges)
    if flags["UNDEREXPOSED"]:
        kernel = cv.getStructuringElement(cv.MORPH_ELLIPSE, (5, 5))
        modelfilter = cv.morphologyEx(modelfilter, cv.MORPH_OPEN, kernel)
    modelcontours, _ = cv.findContours(modelfilter, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)

    # find the biggest model contour (modelC) by area
    if len(modelcontours) == 0:
        modelC = None
        flags["MODEL_CONTOUR_FAILED"] = True
        if log is not None:
            log.write("no model contours found")
    else:
        modelC = max(modelcontours, key=cv.contourArea)

    contour_dict = {"MODEL": modelC, "SHOCK": shockC}

    return contour_dict, flags


def contoursHSV(orig, log=None, minHSVModel=(0, 0, 150), maxHSVModel=(181, 125, 256), minHSVShock=(125, 78, 115), maxHSVShock=(145, 190, 230)):
    """
    Find contours using HSV ranges image.
    Uses the BGR-HSV transformation to increase contrast.

    :param orig: opencv 8bit BGR image
    :param minHSVModel: minimum tuple for HSV range
    :param maxHSVModel: maximum tuple for HSV range
    :param minHSVShock: minimum tuple for HSV range
    :param maxHSVShock: maximum tuple for HSV range
    :returns: model contour, shock contour
    """

    flags = {"MODEL_CONTOUR_FAILED": False, "SHOCK_CONTOUR_FAILED": False}
    # Load an color image in HSV, apply HSV transform again
    hsv_ = cv.cvtColor(orig, cv.COLOR_BGR2HSV)
    hsv = cv.GaussianBlur(hsv_, (5, 5), 0)

    ### Model contours
    modelmask = cv.inRange(hsv, minHSVModel, maxHSVModel)
    modelcontours, _ = cv.findContours(modelmask, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)

    # find the biggest model contour (modelC) by area
    if len(modelcontours) == 0:
        modelC = None
        flags["MODEL_CONTOUR_FAILED"] = True
        if log is not None:
            log.write("no shock contours found")
    else:
        modelC = max(modelcontours, key=cv.contourArea)

    ### Shock contours
    shockmask = cv.inRange(hsv, minHSVShock, maxHSVShock)
    shockcontours, _ = cv.findContours(shockmask, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)

    # find the biggest shock contour (shockC) by area
    if len(shockcontours) == 0:
        shockC = None
        flags["SHOCK_CONTOUR_FAILED"] = True
        if log is not None:
            log.write("no shock contours found")
    else:
        shockC = max(shockcontours, key=cv.contourArea)

    contour_dict = {"MODEL": modelC, "SHOCK": shockC}

    return contour_dict, flags


def contoursGRAY(orig, thresh=150, log=None):
    """
    Detects contours in potentially overexposed images using a global grayscale threshold.

    This function first converts the input image to grayscale and then applies a binary threshold
    to isolate potentially overexposed regions. The function then detects and returns the largest contour
    found, if any, based on the area. In the event no contours are detected, the function logs the failure
    if a logging object is provided.

    Parameters:
    :param orig: An OpenCV BGR image (8-bit depth). This is typically an image that may contain
                 overexposed regions where contours need to be identified.
    :param thresh: Integer value to apply as a threshold for binary segmentation. Defaults to 150.
                   Regions with grayscale values greater than this threshold will be considered as potential contour areas.
    :param log: An optional logging object with a `write` method to capture any logging messages.
                Especially useful to track when contour detection fails. Default is None.

    Returns:
    :return: A tuple containing:
             - `contour_dict`: Dictionary with keys "MODEL" and "SHOCK".
               "MODEL" will contain the largest detected contour or None if not found.
               "SHOCK" is currently always set to None.
             - `flags`: Dictionary indicating success or failure of contour detection.
               Contains keys "SHOCK_CONTOUR_FAILED" (always set to True)
               and "MODEL_CONTOUR_FAILED" (set to True if contour detection fails).

    Example:
        img = cv.imread('overexposed_image.jpg')
        contours, flags = contoursGRAY(img)
        if not flags['MODEL_CONTOUR_FAILED']:
            cv.drawContours(img, [contours['MODEL']], -1, (0, 255, 0), 2)
            cv.imshow('Detected Contour', img)
            cv.waitKey(0)
    """
    flags = {"SHOCK_CONTOUR_FAILED": True, "MODEL_CONTOUR_FAILED": True}
    ### take channel with least saturation
    gray_ = cv.cvtColor(orig, cv.COLOR_BGR2GRAY)

    ### Global grayscale threshold
    gray = cv.GaussianBlur(gray_, (5, 5), 0)
    _, th1 = cv.threshold(gray, thresh, 255, cv.THRESH_BINARY)
    contours, _ = cv.findContours(th1, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)
    if len(contours) != 0:
        # find the biggest contour (c) by the area
        modelC = max(contours, key=cv.contourArea)
        flags["MODEL_CONTOUR_FAILED"] = False
    else:
        if log is not None:
            log.write("no GRAY model contours found at thresh==%i" % thresh)
        modelC = None
        flags["MODEL_CONTOUR_FAILED"] = True

    contour_dict = {"MODEL": modelC, "SHOCK": None}

    return contour_dict, flags


def contoursCNN(orig, model, log=None):
    """
    Find contours using HSV ranges image.
    Uses the BGR-HSV transformation to increase contrast.

    :param orig: opencv 8bit BGR image
    :param model: compiled CNN model
    :returns: model contour, shock contour
    """

    flags = {"MODEL_CONTOUR_FAILED": False, "SHOCK_CONTOUR_FAILED": False}

    ### Apply CNN
    cnnmask = model.predict(orig)

    ### Model contours
    modelmask = (((cnnmask == 1) | (cnnmask == 3)) * 255).astype(np.uint8)
    modelcontours, _ = cv.findContours(modelmask, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)

    # find the biggest model contour (modelC) by area
    if len(modelcontours) == 0:
        modelC = None
        flags["MODEL_CONTOUR_FAILED"] = True
        if log is not None:
            log.write("no shock contours found")
    else:
        modelC = max(modelcontours, key=cv.contourArea)

    ### Shock contours
    shockmask = ((cnnmask == 2) * 255).astype(np.uint8)
    shockcontours, _ = cv.findContours(shockmask, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)

    # find the biggest shock contour (shockC) by area
    if len(shockcontours) == 0:
        shockC = None
        flags["SHOCK_CONTOUR_FAILED"] = True
        if log is not None:
            log.write("no shock contours found")
    else:
        shockC = max(shockcontours, key=cv.contourArea)

    contour_dict = {"MODEL": modelC, "SHOCK": shockC}

    return contour_dict, flags



import os
import unittest
import numpy as np
from arcjetCV.utils.utils import splitfn
from arcjetCV.segmentation.contour.contour import filter_hsv_ranges, getPoints


class TestSplitFn(unittest.TestCase):

    def test_splitfn(self):
        # Test with a standard path
        path, name, ext = splitfn("/user/documents/file.txt")
        self.assertEqual(path, os.path.abspath("/user/documents"))
        self.assertEqual(name, "file")
        self.assertEqual(ext, ".txt")
        
        # Test with no extension
        path, name, ext = splitfn("/user/documents/file")
        self.assertEqual(path, os.path.abspath("/user/documents"))
        self.assertEqual(name, "file")
        self.assertEqual(ext, "")

        # Test with multiple dots in filename
        path, name, ext = splitfn("/user/documents/file.part1.txt")
        self.assertEqual(path, os.path.abspath("/user/documents"))
        self.assertEqual(name, "file.part1")
        self.assertEqual(ext, ".txt")

        # Test with relative path
        path, name, ext = splitfn("documents/file.txt")
        expected_path = os.path.abspath("documents")
        self.assertEqual(path, expected_path)
        self.assertEqual(name, "file")
        self.assertEqual(ext, ".txt")

class TestFilterHsvRanges(unittest.TestCase):

    def setUp(self):
        # Create a simple mock HSV image: 100x100, with a gradient in the hue channel
        self.hsv_image = np.zeros((100, 100, 3), dtype=np.uint8)
        self.hsv_image[:, :, 0] = np.linspace(0, 100, 100, endpoint=False).reshape(-1, 1)  # H gradient
        self.hsv_image[:, :, 1] = 128  # S constant
        self.hsv_image[:, :, 2] = 128  # V constant

    def test_filter_hsv_ranges(self):
        # Define a simple range: 50<H<100, 0<S<255, 0<V<255
        mask = filter_hsv_ranges(self.hsv_image, ([ [50, 0, 0] ], [ [100, 255, 255] ]))

        # Check the output mask: white in the specified range, black elsewhere
        for i in range(100):
            j=50
            if 50 <= i < 100:
                self.assertEqual(mask[i, j], 255)  # White
            else:
                self.assertEqual(mask[i, j], 0)    # Black

        # Test with multiple ranges
        mask_multi = filter_hsv_ranges(self.hsv_image, ([ [10, 0, 0], [95, 0, 0] ], [ [20, 255, 255], [85, 255, 255] ]))
        for i in range(100):
            j=50
            if 10 <= i <= 20 or 95 <= i <= 85:
                self.assertEqual(mask_multi[i, j], 255)  # White
            else:
                self.assertEqual(mask_multi[i, j], 0)    # Black

class TestGetPoints(unittest.TestCase):
    
    def test_right_flow(self):
        # Mock contour for flow_direction="right"
        contour = np.array([[[10, 10]], [[10, 15]], [[10, 20]], [[10, 25]], [[10, 30]]])
        result = getPoints(contour, flow_direction="right")
        
        # Asserting on some expected results. You should adapt these based on your expectations.
        self.assertEqual(result["MODEL_YCENTER"], 20)
        self.assertEqual(result["MODEL_YLOW"], 10)
        self.assertEqual(result["MODEL_CROP_YMAX"], 30)
        self.assertEqual(result["MODEL_RADIUS"], 10)
        self.assertEqual(len(result["MODEL_INTERP_XPOS"]), 5)
    
    def test_left_flow(self):
        # Mock contour for flow_direction="left"
        contour = np.array([[[5, 10]], [[5, 15]], [[5, 20]], [[5, 25]], [[5, 30]]])
        result = getPoints(contour, flow_direction="left")
        
        # Asserting on some expected results
        self.assertEqual(result["MODEL_YCENTER"], 20)
        self.assertEqual(result["MODEL_YLOW"], 10)
        self.assertEqual(result["MODEL_CROP_YMAX"], 30)
        self.assertEqual(result["MODEL_RADIUS"], 10)
        self.assertEqual(len(result["MODEL_INTERP_XPOS"]), 5)
    
    def test_up_flow(self):
        # Mock contour for flow_direction="up"
        contour = np.array([[[10, 5]], [[15, 5]], [[20, 5]], [[25, 5]], [[30, 5]]])
        result = getPoints(contour, flow_direction="up")
        
        # Asserting on some expected results
        self.assertEqual(result["MODEL_YCENTER"], 20)
        self.assertEqual(result["MODEL_YLOW"], 10)
        self.assertEqual(result["MODEL_CROP_YMAX"], 30)
        self.assertEqual(result["MODEL_RADIUS"], 10)
        self.assertEqual(len(result["MODEL_INTERP_XPOS"]), 5)
    
    def test_down_flow(self):
        # Mock contour for flow_direction="down"
        contour = np.array([[[10, 10]], [[15, 10]], [[20, 10]], [[25, 10]], [[30, 10]]])
        result = getPoints(contour, flow_direction="down")
        
        # Asserting on some expected results
        self.assertEqual(result["MODEL_YCENTER"], 20)
        self.assertEqual(result["MODEL_YLOW"], 10)
        self.assertEqual(result["MODEL_CROP_YMAX"], 30)
        self.assertEqual(result["MODEL_RADIUS"], 10)
        self.assertEqual(len(result["MODEL_INTERP_XPOS"]), 5)


if __name__ == "__main__":
    unittest.main()
